"""This module implements the Linear Quadratic Regulator."""

# Mathematics
from numpy.linalg import inv
from numpy import ndarray
from numpy import ones
from scipy.linalg import solve
from scipy.linalg import solve_continuous_are

# Data modelling
from pandas import concat
from pandas import DataFrame


class Lqr:
    """The LQR controller.

    The linear-quadratic-regulator (LQR) is a feedback controller that solves linear-quadratic
    problems at minimum cost. Such problems are defined by linear differential equations and
    quadratic cost functions.

    Examples:
        First, import some helper functions from numpy.

        >>> from numpy import array
        >>> from numpy import eye
        >>> from numpy import vstack

        We then setup the Lqr controller with some control constants.

        >>> controller = Lqr(
        ...     array([[0, 1], [0, 0]]),
        ...     array([0, 1])[:, None],
        ...     array([1, 0])[None, :],
        ...     eye(2),
        ...     array([[1.0]]),
        ...     0.5,
        ... )

        We then specify an initial and desired state.

        >>> initial = vstack([0.0, 0.0])
        >>> desired = vstack([0.0])

        Finally, we retrieve a control signal from the Lqr based on the values we just set.

        >>> signal = controller.control(desired, initial)

    Args:
        A: System matrix (continous time) ``(n, n)``
        B: Input matrix ``(n, 1)``
        C: Output matrix ``(1, n)``
        Q: State cost matrix (pos. semi definite, symmetric) ``(n, n)``
        R: Control cost matrix (pos. definite, symmetric) ``(1, n)``
        dt: Time between measurements
        keep_trace: Whether to store a trace of control signals, states, etc.
        calculate_feed_forward: Whether to compute a feed forward part

    References:
        - https://en.wikipedia.org/wiki/Linear%E2%80%93quadratic_regulator
    """

    # In this context, we reproduce a common PID notation
    # pylint: disable=invalid-name, too-many-arguments

    def __init__(
        self,
        A: ndarray,
        B: ndarray,
        C: ndarray,
        Q: ndarray,
        R: ndarray,
        dt: float,
        keep_trace: bool = False,
        calculate_feed_forward: bool = True,
    ) -> None:  # noqa: E741
        # Dimensionality checks
        assert len(A.shape) == 2 and A.shape[0] == A.shape[1], "Matrix A is not square!"
        assert B.shape[0] == A.shape[0], "Wrong shape for input matrix B!"
        assert C.shape[1] == A.shape[0], "Wrong shape for output matrix C!"

        # Controller specification
        self.A = A
        self.B = B
        self.C = C
        self.Q = Q
        self.R = R
        self.dt = dt

        # Compute controller gain
        # For reference, see here: https://en.wikipedia.org/wiki/Linear%E2%80%93quadratic_regulator
        self.P = solve_continuous_are(self.A, self.B, self.Q, self.R)
        self.K = solve(self.R, self.B.T @ self.P)

        # Calculate static feed forward
        if calculate_feed_forward:
            self.V = inv(-self.C @ inv(self.A - self.B @ self.K) @ self.B)
        else:
            self.V = ones([1, 1])

        # Objects for process tracing
        self.keep_trace = keep_trace
        self.process = DataFrame(
            columns=[
                "desired",
                "state",
                "error",
                "control_signal",
            ]
        )

    def control(self, desired: ndarray, state: ndarray) -> ndarray:
        """Compute the control signal based on LQR controller.

        Args:
            desired: The desired output
            state: The current state

        Returns:
             The control signal
        """

        # Compute errors
        error = desired - self.C @ state

        # Compute feedback and feed forward values
        control_signal: ndarray = -self.K @ state + self.V @ desired

        # Append control step to process trace
        if self.keep_trace:
            new = DataFrame(
                {
                    "desired": (desired.copy(),),
                    "state": (state.copy(),),
                    "error": (error.copy(),),
                    "control_signal": (control_signal.copy(),),
                }
            )
            self.process = concat([self.process, new], ignore_index=True)

        # Return result
        return control_signal

    def reset(self) -> None:
        """Resets the filter's memory, i.e. set the error integral to zero and empty the process trace."""

        self.process = self.process[0:0]
