"""This module contains common base classes for the geospatial objects like polygons, routes and points."""

# Standard library
from abc import ABC
from abc import abstractmethod
from enum import Enum
from enum import IntEnum
from math import pi

# Typing
from typing import Any
from typing import cast
from typing import Dict
from typing import Optional
from typing import Union

# Geospatial
from geojson import dumps
from geojson import Feature


#: The mean earth radius at the equator in meters (taken from
#: `Earth radius (Wikipedia) <https://en.wikipedia.org/wiki/Earth_radius#Mean_radius>`__).
MEAN_EARTH_RADIUS = 6371_008.8

#: The mean earth circumference in meters (derived form :attr:`~MEAN_EARTH_RADIUS`).
MEAN_EARTH_CIRCUMFERENCE = MEAN_EARTH_RADIUS * 2.0 * pi

#: The maximal earth circumference in meters (i.e. at the equator; taken from
#: `Earth's circumference (Wikipedia) <https://en.wikipedia.org/wiki/Earth%27s_circumference>`__).
MAXIMUM_EARTH_CIRCUMFERENCE = 40_075_017.0


class LocationType(IntEnum):

    """Represents what type a location is of.

    Notes:
        The values are set to fixed values such that they can be serialized.
        New members may therefore only be added below, with strictly ascending numbers.
    """

    #: An object of unknown type.
    UNKNOWN = 0

    #: An abstract thing that is used for testing purposes and may not have a correspondence in the real
    #: world. Ships should usually avoid such obstacles too, as they could represent things like
    #: `icebergs <https://de.wikipedia.org/wiki/RMS_Titanic>`__.
    TESTING = 1

    #: A generic obstruction like a buoy, oil rig or special area extracted form a nautical chart.
    OBSTRUCTION = 2

    #: A land mass like an island or continent.
    LAND = 3

    #: Water area that might be considered not sufficiently deep for navigation depending on the context.
    SHALLOW_WATER = 4

    #: Some object representing special weather conditions, like strong winds or just precipitation.
    WEATHER = 5

    #: An object representing other vessels 🚢.
    #: It might have been detected via
    #: `AIS <https://en.wikipedia.org/wiki/Automatic_identification_system>`__.
    SHIP = 6

    @classmethod
    def max_value(cls) -> int:
        """Get the maximum value of all members of this enum."""

        return max(cls)


class Direction(float, Enum):

    """A simple collection of named "compass" bearings 🧭 in degrees for self-documenting code."""

    # pylint: disable=invalid-name
    North = 0.0
    East = 90.0
    South = 180.0
    West = 270.0


class Geospatial(ABC):

    """The common abstract base class for both polar and cartesian geospatial objects.

    See :meth:`~Geospatial.to_geo_json` for hints on how this class can be used for visualizing geometries.

    Args:
        location_type: The type of this polygon
        name: An optional name of this polygon
        identifier: An optional unique identifier for this object, in :math:`[0, 2^{63})`, i.e. 64 signed bits
    """

    def __init__(self, location_type: LocationType, name: Optional[str], identifier: Optional[int]) -> None:
        self.location_type = location_type
        self.name = name
        self.identifier = identifier

        super().__init__()

    @property
    def identifier(self) -> Optional[int]:
        """The numerical identifier of this object.

        Must be `None` or in :math:`[0, 2^{63})`, i.e. 64 signed bits.
        """

        return self._identifier

    @identifier.setter
    def identifier(self, value: Optional[int]) -> None:
        assert value is None or 0 <= value < 2**63, "Identifiers must be in [0, 2**63) or None"

        self._identifier = value

    def to_geo_json(self, indent: Optional[Union[int, str]] = None, **kwargs) -> str:
        """Returns the GeoJSON representation of the geometry embedded into a feature.

        Args:
            indent: the number of levels to indent or ``None`` for compactness (see :func:`json.dumps`)
            kwargs: much like indent, any keyword argument that can be passed to :func:`json.dumps`,
                    like ``allow_nan``, ``sort_keys``, and more

        Returns:
            The GeoJSON representation as a string

        Examples:
            See also: :ref:`geometry-plotting`.

            GeoJSON is a widely used format that can be interpreted by a variety of GIS programs (geo
            information systems). Among them are for example the very simple website
            `geojson.io <https://geojson.io/>`__.
            However, sometimes the geometries are too large to be handled by the web browser.
            Then there are other programs available, like the free open-source tool
            `QGIS (Desktop) <https://www.qgis.org/de/site/>`__. Its even available in the usual Ubuntu
            repositories, so just run ``[sudo] apt install qgis``. later, you can simply copy-pasta it into
            the tool.

            The geojson representation can be obtained like this (using a
            :class:`~pyrate.plan.geometry.location.PolarLocation` just as an example):

            >>> from pyrate.plan.geometry import PolarLocation
            >>> team_room = PolarLocation(latitude=49.878091, longitude=8.654052)
            >>> print(team_room.to_geo_json(indent=4))
            {
                "type": "Feature",
                "geometry": {
                    "type": "Point",
                    "coordinates": [
                        8.654052,
                        49.878091
                    ]
                },
                "properties": {}
            }

        See also:
            - `GeoJSON on Wikipedia <https://en.wikipedia.org/wiki/GeoJSON>`__
            - `geojson.io <https://geojson.io/>`__
            - `QGIS (Desktop) <https://www.qgis.org/de/site/>`__
        """

        # this relies on the inheriting instance to provide __geo_interface__ property/attribute
        return cast(str, dumps(Feature(geometry=self), indent=indent, **kwargs))

    @property
    @abstractmethod
    def __geo_interface__(self) -> Dict[str, Any]:
        raise NotImplementedError()

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Geospatial)
            and self.location_type == other.location_type
            and self.name == other.name
            and self.identifier == other.identifier
        )

    @property
    def _repr_extras(self) -> str:
        """Create a string representation of the three extra attributes for use in :meth:`~__repr__`.

        Examples:
            The output is suited to be directly inlucded before the final closing bracet of a typical
            implementation of ``__repr__()``:

            >>> from pyrate.plan.geometry import PolarLocation
            >>> PolarLocation(0, 0)._repr_extras
            ''
            >>> PolarLocation(0, 0, location_type=LocationType.UNKNOWN, name=None)._repr_extras
            ''
            >>> PolarLocation(0, 0, name="")._repr_extras
            ', name=""'
            >>> PolarLocation(0, 0, location_type=LocationType.SHIP, identifier=12)._repr_extras
            ', location_type=LocationType.SHIP, identifier=12'

            The class :class:`pyrate.plan.geometry.location.PolarLocation` was only chosen as an example.

        Returns:
            The extra arguments in the syntax of keyword arguments, as is common for :meth:`~__repr__`.
        """

        result = ""

        if self.location_type != LocationType.UNKNOWN:
            result += f", location_type=LocationType.{self.location_type.name}"
        if self.name is not None:
            result += f', name="{self.name}"'
        if self.identifier is not None:
            result += f", identifier={self.identifier}"

        return result

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError()
