"""This module implements abstractions for geospatial routes (line strings) in WGS84 and local coordinate
frames.

Two routes are ``==`` if and only if they are equal according to ``equals_exact()``.
"""

# Python standard library
from copy import deepcopy
from math import cos
from math import radians
from math import sin

# Typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

# Data modelling
from numpy import array
from numpy import isfinite
from numpy import ndarray
from shapely.affinity import translate
from shapely.geometry import LineString

# Geospatial basis
from .geospatial import Geospatial
from .geospatial import LocationType

# Geospatial helpers
from .helpers import mean_coordinate
from .helpers import translate_numpy

# Location representation
from .location import CartesianLocation
from .location import PolarLocation


class PolarRoute(Geospatial):

    """A route (line string) based on WGS84 coordinates.

    Note:
        This class does not yet support simplification as it was not required so far.

    Args:
        locations: The two or more points that make up this route; see :attr:`~.locations`
        location_type: The type of this polygon
        name: An optional name of this polygon
        identifier: The route's optional unique identifier, in :math:`[0, 2**63)`, i.e. 64 signed bits
    """

    def __init__(
        self,
        locations: List[PolarLocation],
        location_type: LocationType = LocationType.UNKNOWN,
        name: Optional[str] = None,
        identifier: Optional[int] = None,
    ) -> None:
        # Type hints
        self._locations: List[PolarLocation]

        # Attributes setup
        self.locations = locations
        self._mean_location: Optional[PolarLocation] = None

        super().__init__(location_type=location_type, name=name, identifier=identifier)

        # See Shapely issue
        if self.length(approximate=True) < 1e-9:
            raise ValueError(f"(Nearly) zero-length line strings are not allowed by Shapely; got {locations}")

    @property
    def locations(self) -> List[PolarLocation]:
        """The points that make up this route.

        Getter:
            At least two points are returned.

        Setter:
            Raises an :class:`AssertionError` if less than two points are given.
        """

        return self._locations

    @locations.setter
    def locations(self, locations: List[PolarLocation]) -> None:
        assert len(locations) >= 2, "a route must contain at least two points"
        self._locations = locations
        self._mean_location = None

    def distance_to_vertices(self, location: PolarLocation, approximate: bool = False) -> float:
        """Computes the distance of the given ``location`` to the nearest vertex of this route.

        Args:
            location: The location to compute the distance from
            approximate: Whether to use a less precise, faster method or not
        """

        return min([location.distance(loc, approximate) for loc in self.locations])

    def length(self, approximate: bool = False) -> float:
        """Compute the length of this route from start to end.

        Args:
            approximate: Whether to use a less precise, faster method or not
        """

        return sum([a.distance(b, approximate) for (a, b) in zip(self.locations[:-1], self.locations[1:])])

    def to_cartesian(self, origin: PolarLocation) -> "CartesianRoute":
        """Projects this route to a cartesian representation according to the given global reference.

        Args:
            origin: The reference by which to project onto the local tangent plane

        Returns:
            The cartesian representation of this route with the given reference point being set
        """

        # convert to cartesian
        coordinates = self.to_numpy()
        coordinates[:, 0], coordinates[:, 1] = origin.projection(coordinates[:, 0], coordinates[:, 1])

        return CartesianRoute.from_numpy(
            coordinates,
            origin=origin,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
        )

    def to_numpy(self) -> ndarray:
        """Converts the coordinates defining this route into a :class:`numpy.ndarray`.

        Returns:
            An array with shape ``(number of locations, 2)``, where each location is represented by a pair of
            ``(longitude, latitude)``, each in degrees.

        See Also:
            :meth:`~from_numpy`
        """

        return array(
            [(location.longitude, location.latitude) for location in self.locations],
            dtype="float64",
            order="C",
        )

    @classmethod
    def from_numpy(cls, data: ndarray, *args, **kwargs) -> "PolarRoute":
        """Create a polar route from a numpy representation.

        Args:
            data: An array with shape ``(number of locations, 2)``, where each location is represented by a
                   pair of ``(longitude, latitude)``, each in degrees.
            *args: Positional arguments to be passed to :class:`~PolarRoute`
            **kwargs: Keyword arguments to be passed to :class:`~PolarRoute`

        Returns:
            The polar route created from the given coordinates and other parameters

        Raises:
            AssertionError: If the shape of ``data`` is invalid or contains non-finite values

        See Also:
            :meth:`~to_numpy`
        """

        assert len(data.shape) == 2
        assert data.shape[1] == 2
        assert isfinite(data).all(), "Invalid values in CartesianRoute.from_numpy()"

        return cls([PolarLocation(latitude=lat, longitude=lon) for (lon, lat) in data], *args, **kwargs)

    def translate(self, direction: float, distance: float) -> Tuple["PolarRoute", ndarray]:
        """Translates this location and returns the new route and back-azimuth.

        See :func:`pyrate.plan.geometry.helpers.translate_floats` for details.
        """

        new_coordinates, back_azimuth_array = translate_numpy(self.to_numpy(), direction, distance)

        new_route = PolarRoute.from_numpy(
            new_coordinates,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
        )

        return new_route, back_azimuth_array

    @property
    def mean(self) -> PolarLocation:
        """Computes a reasonable mean location of the route, if possible. The result is cached.

        Raises:
            ValueError: If no meaningful mean (of the longitudes) can be computed.
                See :func:`pyrate.plan.geometry.helpers.mean_angle`.
        """

        if self._mean_location is None:
            coordinates = self.to_numpy()
            latitude, longitude = mean_coordinate(latitudes=coordinates[:, 1], longitudes=coordinates[:, 0])
            name = f"{self.name} - mean" if self.name else "mean"
            self._mean_location = PolarLocation(latitude, longitude, name=name)

        return self._mean_location

    @property
    def __geo_interface__(self) -> Dict[str, Any]:
        return {
            "type": "LineString",
            "coordinates": [(location.longitude, location.latitude) for location in self.locations],
        }

    def __eq__(self, other: Any) -> bool:
        return self.equals_exact(other, tolerance=0.0)

    def equals(self, other: Any) -> bool:  # pylint: disable=missing-function-docstring
        return (
            isinstance(other, PolarRoute)
            and self.to_cartesian(self.mean).equals(other.to_cartesian(self.mean))
            and Geospatial.__eq__(self, other)
        )

    equals.__doc__ = PolarLocation.equals.__doc__

    def equals_exact(self, other: Any, tolerance: float) -> bool:
        # pylint: disable=missing-function-docstring

        return (
            isinstance(other, PolarRoute)
            and self.to_cartesian(self.mean).equals_exact(other.to_cartesian(self.mean), tolerance)
            and Geospatial.__eq__(self, other)
        )

    equals_exact.__doc__ = PolarLocation.equals_exact.__doc__

    def __repr__(self) -> str:
        locations = ", ".join(str(loc) for loc in self.locations)
        return f"PolarRoute(locations=[{locations}]{self._repr_extras})"


class CartesianRoute(Geospatial, LineString):

    """A cartesian route (line string) in local coordinates, optionally with a global reference point.

    Note:
        For the sake of simplicity and performance, this class does not store the given
        :class:`~pyrate.plan.geometry.location.CartesianLocation` instances directly,
        but only their coordinates.
        Thus, when reading back attributes like ``origin``, ``name``, etc. of the locations they are derived
        from the route instance and not from the individual locations.

    Args:
        locations: The list of two or more locations that this shape consists of; see :attr:`~locations`
        location_type: The type of this route
        name: The name of this route
        identifier: The route's optional unique identifier, in :math:`[0, 2**63)`, i.e. 64 signed bits
        origin: A reference that can be used to project this cartesian representation (back)
            into a polar one
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        locations: List[CartesianLocation],
        location_type: LocationType = LocationType.UNKNOWN,
        name: Optional[str] = None,
        identifier: Optional[int] = None,
        origin: Optional[PolarLocation] = None,
    ) -> None:
        # Store attributes
        self.origin = origin

        if isinstance(locations, list):
            LineString.__init__(self, [location.coords[0] for location in locations])
        else:
            # this is required for an efficient implementation of CartesianRoute.from_numpy
            # we do not add this possibility to the type signature to make people use from_numpy().
            LineString.__init__(self, locations)

        Geospatial.__init__(self, location_type=location_type, name=name, identifier=identifier)

        # See Shapely issue
        if self.length < 1e-9:
            raise ValueError(f"(Nearly) zero-length line strings are not allowed by Shapely; got {locations}")

    @property
    def locations(self) -> List[CartesianLocation]:
        """Get the locations of this route. See the class description for caveats."""

        return [
            CartesianLocation(
                x,
                y,
                origin=self.origin,
                location_type=self.location_type,
                name=self.name,
                identifier=self.identifier,
            )
            for (x, y) in self.coords
        ]

    def to_polar(self, origin: Optional[PolarLocation] = None) -> PolarRoute:
        """Computes the polar representation of this route.

        Args:
            origin: The global reference to be used for back-projection, must be set if and only if
                :attr:`~origin` is ``None``

        Returns:
            The global, polar representation of this route
        """

        if origin is None:
            if self.origin is None:
                raise ValueError("need to give an explicit origin when the instance does not have one")
            origin = self.origin
        elif self.origin is not None:
            raise ValueError("provided an explicit origin while the instance already has one")

        # convert to cartesian
        coordinates = self.to_numpy()
        coordinates[:, 0], coordinates[:, 1] = origin.projection(
            coordinates[:, 0], coordinates[:, 1], inverse=True
        )

        return PolarRoute.from_numpy(
            coordinates, location_type=self.location_type, name=self.name, identifier=self.identifier
        )

    def to_numpy(self) -> ndarray:
        """Converts the coordinates defining this route into a :class:`numpy.ndarray`.

        Returns:
            An array with shape ``(number of locations, 2)``, where each location is represented by a pair of
            ``(longitude, latitude)``, each in degrees.

        See Also:
            :meth:`~from_numpy`
        """

        return array(self.coords, dtype="float64", order="C")

    @classmethod
    def from_numpy(cls, data: ndarray, *args, **kwargs) -> "CartesianRoute":
        """Create a cartesian route from a numpy representation.

        Args:
            data: An array with shape ``(number of locations, 2)``, where each location is represented by a
                  pair of ``(longitude, latitude)``, each in degrees.
            *args: positional arguments to be passed to :class:`~CartesianRoute`
            **kwargs: keyword arguments to be passed to :class:`~CartesianRoute`

        Returns:
            The cartesian route created from the given coordinates and other parameters

        Raises:
            AssertionError: If the shape of ``data`` is invalid or contains non-finite values

        See Also:
            :meth:`~to_numpy`
        """

        assert len(data.shape) == 2
        assert data.shape[1] == 2
        assert isfinite(data).all(), "Invalid values in PolarRoute.from_numpy()"

        return cls(data, *args, **kwargs)  # type: ignore

    @classmethod
    def from_shapely(cls, line_string: LineString, *args, **kwargs) -> "CartesianRoute":
        """Create a cartesian route from a shapely line string.

        Args:
            line_string: A shapely line_string
            *args: Positional arguments to be passed to :class:`~CartesianRoute`
            **kwargs: Keyword arguments to be passed to :class:`~CartesianRoute`

        Returns:
            The cartesian route created from the given geometry and other parameters
        """
        return cls.from_numpy(array(line_string.xy).T, *args, **kwargs)

    def translate(self, direction: float, distance: float) -> Tuple["CartesianRoute", ndarray]:
        """Translates this route.

        Args:
            direction: The direction angle in degrees (``0`` is north, clockwise)
            distance: The distance to translate bin meters

        Returns:
            The translated route and the translation vector ``(x_offset, y_offset)`` in meters
            that can be used to reconstruct the original route
        """

        x_offset = sin(radians(direction)) * distance
        y_offset = cos(radians(direction)) * distance

        return (
            CartesianRoute.from_shapely(
                translate(LineString(self.to_numpy()), xoff=x_offset, yoff=y_offset),
                location_type=self.location_type,
                name=self.name,
                identifier=self.identifier,
                origin=self.origin,
            ),
            array([-x_offset, -y_offset]),
        )

    @property
    def __geo_interface__(self) -> Dict[str, Any]:
        return {"type": "LineString", "coordinates": self.coords}

    def __copy__(self) -> "CartesianRoute":
        return CartesianRoute(
            locations=self.locations,
            location_type=self.location_type,
            name=self.name,
            identifier=self.identifier,
            origin=self.origin,
        )

    def __deepcopy__(self, memodict: Dict) -> "CartesianRoute":
        return CartesianRoute(
            locations=deepcopy(self.locations, memodict),
            location_type=deepcopy(self.location_type, memodict),
            name=deepcopy(self.name, memodict),
            identifier=deepcopy(self.identifier, memodict),
            origin=deepcopy(self.origin, memodict),
        )

    def __eq__(self, other: Any) -> bool:
        return self.equals_exact(other, tolerance=0.0)

    # Inherits the docstring
    def equals(self, other: Any) -> bool:  # pylint: disable=missing-function-docstring
        return (
            isinstance(other, CartesianRoute)
            and LineString.equals(self, other)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    # Inherits the docstring
    def equals_exact(  # pylint: disable=missing-function-docstring
        self, other: Any, tolerance: float
    ) -> bool:
        return (
            isinstance(other, CartesianRoute)
            and LineString.equals_exact(self, other, tolerance)
            and Geospatial.__eq__(self, other)
            and self.origin == other.origin
        )

    def __repr__(self) -> str:
        origin = f", origin={self.origin}" if self.origin is not None else ""
        locations = ", ".join(f"({x}, {y})" for x, y in self.coords)

        return f"CartesianRoute(locations=[{locations}]{origin}{self._repr_extras})"

    def __str__(self) -> str:
        # this is required to override shapely.geometry.LineString.__str__()
        return self.__repr__()
