"""
Creates a grid on a globe with vertices and edges. Assumes that the earth is a sphere.


Examples:
    The usual approach it to first determine the maximum allowed distance between two nodes on the graph.
    Note, that small distances (e.g. of less than 100km) might take a while to compute and can create very
    large graphs. See the script
    :ref:`earth_graph_frequency_statistics.py <script-earth_graph_frequency_statistics-example>`
    for details on the performance and size of the output.

    >>> maximum_node_distance = 100_000  # in meters

    Then, the minimum required ``frequency`` can be computed from that, which is an integer value determining
    the granularity of the graph.
    Higher frequencies result in finer graphs.

    >>> frequency = min_required_frequency(maximum_node_distance, in_meters=True)

    We could have also passed the angular distance in radians (by setting ``in_meters=False``).
    Alternatively, we can now compute the actual angular distance and great-circle distance in meters from the
    frequency that we know have.
    It is in general less (or equal) than the ``maximum_node_distance``, as the ``frequency`` only allows for
    integer steps in the geranularity.

    >>> angular_distance_for(frequency)  # doctest: +ELLIPSIS
    0.01559...
    >>> actual_node_distance = great_circle_distance_distance_for(frequency)  # again, in meters
    >>> actual_node_distance  # doctest: +ELLIPSIS
    99347.242...
    >>> actual_node_distance <= maximum_node_distance
    True

    Now, we can finally generate the :class:`~pyrate.plan.graph.geo_graph.GeoNavigationGraph`.
    If we wanted to have some progress messsages printed, we would pass ``print_status=True``.

    >>> graph = create_earth_graph(frequency)
    >>> len(graph)  # the number of nodes
    50412
    >>> graph.num_edges
    151230

    Furthermore, the ``graph`` has some specific attributes set, which result from the icoshedron subdivision
    approach of the algorithm.
    These allow for certain optimizations and more convenience when applying algorithms as they do not have
    to be passed explicictly to other functions.

    >>> graph.node_radius * 2 == actual_node_distance
    True
    >>> graph.max_neighbors == 6
    True

Visualization
-------------

The following visualization shows how the vertices of an earth graph are spread when plotted using the
`mercator projection <https://en.wikipedia.org/wiki/Mercator_projection>`_.
The vertices and their area near the equator are very evenly spaced.
However, their positions and shapes get very distorted at high latitudes (i.e. near the north and south
poles) `due to the projection <https://en.wikipedia.org/wiki/Mercator_projection#Distortion_of_sizes>`_.

.. image:: vertices_distribution_mercator.png
  :alt: visualization of the vertices of an earth graph, distorted by the mercator projection

The following plot illustrates the area of influence/responsibility around each vertex.
Again, notice how the quite evenly spaced vertices are very distorted in this projection at high latitudes.
The visualization was obtained by computing the fraction of navigable area within the nodes' vicinity.
Navigability was simply determined by the land being below sea level as a rough approximation.

.. image:: vertices_area_of_influence.png
  :alt: visualization of the area of influence/responsibility of all vertices of an earth graph,
        distorted by the mercator projection;
        obtained by computing the fraction of navigable area within the nodes vicinity

Note:
    Some methods require the `Antiprism <https://www.antiprism.com>`_ software package (version 0.26+).
    There is a PPA for it available `here <https://launchpad.net/~antiprism/+archive/ubuntu/ppa>`_.
    Use ``add-apt-repository ppa:antiprism/ppa && apt install antiprism`` on Ubuntu.
"""

# Standard library
from math import ceil
from math import degrees
import subprocess
from warnings import warn

# Typing
from typing import Tuple

# Scientific
from numpy import compress
from numpy import empty
from numpy import empty_like
from numpy import float64
from numpy import genfromtxt
from numpy import maximum
from numpy import minimum
from numpy import ndarray
from numpy import uint32
from numpy import unique

# Geometry helpers
from pyrate.plan.geometry.helpers import cartesian_to_spherical
from pyrate.plan.geometry.helpers import meters2rad
from pyrate.plan.geometry.helpers import rad2meters

# Graph implementation
from pyrate.plan.graph import GeoNavigationGraph


def create_earth_graph(frequency: int, print_status: bool = False) -> GeoNavigationGraph:
    """Returns roughly equally spaced points on the earth, creating an *icoshpere* 🌐.

    This basically works by constructing a geodesic polyhedron based on an icosahedron as a starting point,
    dividing it (with *Class I*) as much as required by the desired distance and then projecting it onto a
    sphere. The following image visualizes the process for the case ``frequency = 6``:

    .. image:: https://upload.wikimedia.org/wikipedia/commons/f/ff/Geodesic_icosahedral_polyhedron_example.png
      :alt:
        Illustration of icosahedron subdivision: (1) create a regular icosahedron, (2) perform 6-frequency
        subdivision of all faces and (3) project all vertices on a sphere;
        Licensed under *CC BY-SA 4.0*: Created by *Tomruen* and provided on
        `Wikipedia <https://commons.wikimedia.org/wiki/File:Geodesic_icosahedral_polyhedron_example.png>`_.

    The implementation is mostly single-threaded. See the script
    :ref:`earth_graph_frequency_statistics.py <script-earth_graph_frequency_statistics-example>`
    for details on the performance and size of the output.

    *Class I* determines the way that the icosahedron is sliced (see
    `Wikipedia: Geodesic notation <https://en.wikipedia.org/wiki/Geodesic_polyhedron#Geodesic_notation>`_).
    It was chosen since it is apparently used quite often to create regular grids
    (according to `Wikipedia <https://en.wikipedia.org/wiki/Geodesic_grid#Construction>`_)
    and since if appears to be very regular (see
    `this <https://www.antiprism.com/examples/200_programs/650_geodesic/imagelist.html>`_ page in the
    Antiprism documentation for a visualization).

    References:
        - https://en.wikipedia.org/wiki/Geodesic_grid#Construction
        - https://en.wikipedia.org/wiki/Geodesic_polyhedron
        - https://people.sc.fsu.edu/~jburkardt/presentations/sphere_grid_2014_fsu.pdf

    Further ideas:
        - One could also use `Goldberg polyhedra <https://en.wikipedia.org/wiki/Goldberg_polyhedron>`_,
          as those are the `duals <https://en.wikipedia.org/wiki/Dual_polygon>`_ to the geodesic spheres used
          in this implementation and should also work.
        - Alternatively, one could also use the already implemented
          `Quaternary Triangular Meshes <https://github.com/paulojraposo/QTM>`_

    Args:
        frequency: The number of subdivisions per icosahedron edge.
                   Keep in mind that high frequencies could be computationally costly.
        print_status: If set to ``True``, print human-readable status messages about the progress.

    Returns:
        A graph covering the entire earth
    """
    angular_distance = angular_distance_for(frequency)
    distance_meters = great_circle_distance_distance_for(frequency)
    if print_status:
        print(f"creating an earth grid with a point distance of at most {distance_meters / 1000:.3f} km")
        print(f"dividing each edge of the base icosahedron into {frequency} parts")
        print(f"the angular distance of vertices will be ~{degrees(angular_distance):.6f}°")

    if 10_000 < distance_meters < 25_000:  # pragma: no cover
        warn("this might take a while", ResourceWarning)
    elif distance_meters <= 10_000:  # pragma: no cover
        warn("this might take *very* long", ResourceWarning)

    # this defines how to slice the edges/triangles
    polyhedron_class = "1"

    if print_status:
        print('calling Antiprisim\'s "geodesic"')

    # check the geodesic/Antiprism version
    _assert_antiprism_is_installed()

    command = f"geodesic -M s -c {polyhedron_class} -f {frequency} ico"
    # check_output raises an Error on a non-zero exit code
    # use ASCII encoding since the resulting file will not contain any Unicode text
    output = subprocess.check_output(command.split(), encoding="ascii")

    if print_status:
        print("parsing the resulting OFF file")
    latitudes, longitudes, edges = _parse_off_file(output)

    if print_status:
        print("finished earth grid generation")

    return GeoNavigationGraph.from_coordinates_radians(
        latitudes=latitudes,
        longitudes=longitudes,
        edges=edges,
        max_neighbors=6,
        node_radius=distance_meters / 2,
    )


#: The approximate angle between two edges on an icosahedron, in radians, about 63.4°
_ALPHA = 1.1071487177940905030170654601785
# calculation:
#     (note: we use latitude is in [-pi/2, +pi/2], longitude is in [-pi, +pi])
#     take two edges in spherical coordinates,
#     see https://en.wikipedia.org/wiki/Regular_icosahedron#Spherical_coordinates
#     (in the link, other coordinates are used!)
#     we choose A=(lat_a, lon_a)=(pi/2, 0) and
#     B=(lat_b, lon_b)=(arctan(1/2), 0) for simplicity
#     then the angle between them is given by
#     alpha = lat_a - lat_b = pi/2 - arctan(1/2)
#     result: https://www.wolframalpha.com/input/?i=pi%2F2+-+arctan%281%2F2%29


def min_required_frequency(desired_distance: float, in_meters: bool) -> int:
    """Compute the minimum frequency to reach the ``desired_distance`` by icosahedron subdivision.

    Here, the frequency is the number of cuts to make on an edge of a polyhedron.
    Higher frequencies result in finer graphs.

    Args:
        desired_distance: The maximum distance that two neighboring nodes may be apart.
                          Must be a strictly positive number.
                          If ``in_meters`` is ``True`` in meters of the great-circle distance, else the
                          angular distance in radians.
        in_meters: Interpret ``desired_distance`` as meters instead of as radians

    Returns:
        The minimum frequency to reach the ``desired_distance``, at least ``1``
    """
    assert desired_distance > 0, "the desired_angular_distance must be positive"

    if in_meters:
        desired_angular_distance = meters2rad(desired_distance)
    else:
        desired_angular_distance = desired_distance

    # calculate the number of slices per edge (=the frequency) by simple division:
    frequency = _ALPHA / desired_angular_distance

    # if the distance is too big, we simply do not divide the edges at all
    frequency = max(frequency, 1.0)

    # then we need to round: we round up since we would rather have
    # more edges than too few of them
    return int(ceil(frequency))


def great_circle_distance_distance_for(frequency: int) -> float:
    """The great-circle distance that subdivision with the frequency will result in.

    Args:
        frequency: The frequency of the subdivision, at least ``1``

    Returns:
        The great-circle distance that the frequency will result in, in meters
    """
    return rad2meters(angular_distance_for(frequency))


def angular_distance_for(frequency: int) -> float:
    """The angular distance that subdivision with the frequency will result in.

    Args:
        frequency: The frequency of the subdivision, at least ``1``

    Returns:
        The angular distance that the frequency will result in, in radians
    """
    assert frequency >= 1, "the frequency must be at least one"
    return _ALPHA / frequency


_ANTIPRISM_REQUIRED_VERSION = (0, 26)
#: The minimum required Antiprism version


def _assert_antiprism_is_installed() -> None:
    """Raises an exception if *Antiprism* (with the geodesic tool) in not installed in the required version.

    Raises:
        :class:`AssertionError`: If the *Antiprism* version is insufficient
    """
    try:
        version = subprocess.check_output(["geodesic", "--version"], encoding="utf8").split(" ", 3)[2]
    except FileNotFoundError as error:  # pragma: no cover
        raise AssertionError(
            'Could not call tool "geodesic" from Antiprism, is it installed? (See installation instructions.)'
        ) from error

    assert tuple(int(v) for v in version.split(".")) >= _ANTIPRISM_REQUIRED_VERSION, (
        f'tool "geodesic" from Antiprism version >= {_ANTIPRISM_REQUIRED_VERSION} is required, '
        f"but you have version {version}!"
    )


def _parse_off_file(source_text: str) -> Tuple[ndarray, ndarray, ndarray]:
    """Parses an Antiprism OFF file and return the result in spherical coordinates.

    Warnings:
        Assumes that the point :math:`(0, 0, 0)` is not present and that all faces
        are triangles or "polygons" with fewer vertices.

    Warnings:
        This is only meant to parse OFF files produced by *Antiprism*.
        These are not standard OFF files as described
        `here <https://www.antiprism.com/programs/off_format.html>`_!

    Args:
        source_text: The raw file content to be parsed

    Returns:
        all vertices' points (like returned by :meth:`~off_handler.cartesian_to_spherical`)
        as well as a list of all edges, each consisting of zero-based indices
        of the endpoints from the first argument.
    """
    # split
    source = source_text.splitlines()
    assert len(source) >= 2, "OFF file must have at least two lines"

    # check header
    assert source[0] == "OFF", 'file does not start with "OFF"'

    # get size of file
    # note: num_edges is usually not set to a correct value, so we ignore the last value
    num_vertices, num_faces, _ = map(int, source[1].split())

    # get the vertices
    points = genfromtxt(source[2:], max_rows=num_vertices, dtype=float64)
    latitudes, longitudes = cartesian_to_spherical(points)

    # get faces
    faces = genfromtxt(source[2 + num_vertices :], max_rows=num_faces, usecols=(0, 1, 2, 3), dtype=uint32)
    triangles = compress(faces[:, 0] == 3, faces[:, 1:4], axis=0)
    del faces  # free this memory
    count = len(triangles)

    # now we want to transform each triangle into three edges
    edges = empty([count * 3, 2], dtype=uint32)
    edges[0:count, :] = triangles[:, (0, 1)]
    edges[count : 2 * count, :] = triangles[:, (1, 2)]
    edges[2 * count : 3 * count, :] = triangles[:, (0, 2)]

    # then we filter out duplicates or wrong values
    # sort the IDs in each row in ascending order, to find duplicates since the  graph is directed
    # one could also use `np.sort`
    sorted_edges = empty_like(edges)
    sorted_edges[:, 0] = minimum(edges[:, 0], edges[:, 1])
    sorted_edges[:, 1] = maximum(edges[:, 0], edges[:, 1])
    edges = unique(sorted_edges, axis=0)

    return latitudes, longitudes, edges
