"""This module provides geo-referenced navigation graphs."""

# Typing
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

# Scientific
import numpy
from numpy import degrees
from numpy import ndarray
from numpy import radians

# Scientific
from pandas import DataFrame
from pandas import Series

# Progress bars
from tqdm import tqdm

# Own typing
from pyrate.common.raster_datasets import BaseTransformer
from pyrate.plan.graph.graph import NavigationGraph


class GeoNavigationGraph(NavigationGraph):
    """An undirected navigation graph specifically for geo-referenced graphs.

    It is similar to the more generic :class:`~pyrate.plan.graph.graph.NavigationGraph` but ensures that the
    property dataframe always contains columns `Latitude (radians)` and `Longitude (radians)`.
    Not providing these when creating the graph will result in an :class:`AssertionError`.

    This class also useful methods for adding new properties and plotting the graph.

    Examples:
        This creates a very simple graph with two connected nodes at *Darmstadt* and *Griesheim*.

        >>> import numpy
        >>> nodes = DataFrame(data={'Latitude (radians)':  numpy.radians([49.872222, 49.863889]), \
                                    'Longitude (radians)': numpy.radians([ 8.652778,  8.563889])})
        >>> edges = numpy.array([[0, 1], ])
        >>> graph = GeoNavigationGraph(nodes, edges)
        >>> graph.neighbors
        array([[1],
               [0]], dtype=int32)
        >>> graph.latitudes_degrees
        0    49.872222
        1    49.863889
        Name: Latitude (radians), dtype: float64

        Alternatively, such a graph can be created using `GeoNavigationGraph.from_coordinates_*`

        >>> same_graph = GeoNavigationGraph.from_coordinates_degrees( \
            latitudes=[49.872222, 49.863889], longitudes=[ 8.652778,  8.563889], edges=edges)
        >>> graph == same_graph
        True

    Args:
        nodes: See :class:`~pyrate.plan.graph.graph.NavigationGraph`.
               This must contain columns ``"Latitude (radians)"`` (with values in :math:`[-π/2, +π/2]`) and
               ``"Longitude (radians)"`` (with values in :math:`[-π, +π)`).
        edges: See :class:`~pyrate.plan.graph.graph.NavigationGraph`.
        neighbours: See :class:`~pyrate.plan.graph.graph.NavigationGraph`.
        max_neighbors: See :class:`~pyrate.plan.graph.graph.NavigationGraph`.
        node_radius: The radius around each node of the area on the globe that it should represent, in meters,
                     non-negative. It can be interpreded as the radius of influence or of responsibility.
                     It may be an array of shape ``(num_nodes, )`` or a single scalar if the radius is uniform
                     across all nodes.
                     Setting this allows it to be omitted in some methods of this class, like in
                     :meth:`~append_property`.
    """

    def __init__(self, *args, node_radius: Optional[Union[float, ndarray]] = None, **kwargs):
        super().__init__(*args, **kwargs)

        assert node_radius is None or numpy.all(node_radius >= 0)
        self.node_radius = node_radius

        assert "Latitude (radians)" in self.nodes, 'column "Latitude (radians)" missing'
        assert "Longitude (radians)" in self.nodes, 'column "Longitude (radians)" missing'

    @classmethod
    def from_coordinates_radians(
        cls,
        latitudes: ndarray,
        longitudes: ndarray,
        node_properties: Optional[DataFrame] = None,
        **kwargs: Any,
    ) -> "GeoNavigationGraph":
        """Creates a new geo-referenced navigation graph from the given coordinates and node properties.

        The same as the constructor of :class:`GeoNavigationGraph`, except that that latitude, longitude and
        properties of the nodes can be given separately. For clarity, everything should be passes as
        keyword arguments.

        Args:
            latitudes: The latitudes of all nodes in radians in :math:`[-π/2, +π/2]`
            longitudes: The longitudes of all nodes in radians in :math:`[-π, +π)`
            node_properties: The properties of all nodes (will be modified if not set to ``None``)
            kwargs**: passed to the constructor of :class:`GeoNavigationGraph`

        Returns:
            A newly created graph
        """
        if node_properties is None:
            node_properties = DataFrame()  # create an empty one

        node_properties["Latitude (radians)"] = latitudes
        node_properties["Longitude (radians)"] = longitudes

        assert "nodes" not in kwargs, (
            "do not pass nodes, instead explicitly set them via latitudes, "
            "longitudes and node_properties or directly use the constructor instead"
        )

        return cls(node_properties, **kwargs)

    @classmethod
    def from_coordinates_degrees(
        cls, latitudes: ndarray, longitudes: ndarray, **kwargs: Any
    ) -> "GeoNavigationGraph":
        """The same as :func:`~from_coordinates_radians` except that the coordinates are in degrees.

        For clarity, everything should be passes as keyword arguments.

        Args:
            latitudes: The latitudes of all nodes in degrees in :math:`[-90, +90]`
            longitudes: The latitudes of all nodes in degrees in :math:`[-180, +180)`
            kwargs**: passed to :func:`~from_coordinates_radians`
        """
        return GeoNavigationGraph.from_coordinates_radians(radians(latitudes), radians(longitudes), **kwargs)

    @staticmethod
    def _serialized_attributes() -> List[str]:
        """The list of attributes that shall be (de)serialized (on top of the nodes and edges)."""
        return NavigationGraph._serialized_attributes() + ["node_radius"]

    @property
    def latitudes_radians(self) -> Series:
        """The latitudes of all nodes in radians in :math:`[-π/2, +π/2]`."""
        return self.nodes["Latitude (radians)"]

    @property
    def longitudes_radians(self) -> Series:
        """The longitudes of all nodes in radians in :math:`[-π, +π)`."""
        return self.nodes["Longitude (radians)"]

    @property
    def latitudes_degrees(self) -> Series:
        """The latitudes of all nodes in degrees in :math:`[-90, +90]`."""
        return degrees(self.latitudes_radians)

    @property
    def longitudes_degrees(self) -> Series:
        """The longitudes of all nodes in degrees in :math:`[-180, +180)`."""
        return degrees(self.longitudes_radians)

    @property
    def node_properties(self) -> DataFrame:
        """The properties of all nodes as a view (as opposed to a copy).

        This is the same as :attr:`~nodes`, but without the latitude and longitude values.
        """
        return self.nodes.drop(columns=["Latitude (radians)", "Longitude (radians)"])

    def clear_node_properties(self) -> None:
        """Deletes all properties but retains the coordinate values."""
        self.nodes = self.nodes[["Latitude (radians)", "Longitude (radians)"]]

    def append_property(
        self,
        transformer: BaseTransformer,
        node_radius: Optional[Union[float, ndarray]] = None,
        show_progress: bool = False,
    ) -> None:
        """Append the properties given by the transformer.

        The name and data type are taken from the given ``transformer``.

        Args:
            transformer: The dimension/property that shall be queried for each node
            node_radius: The radius around each node of the area on the globe that it should represent,
                         in meters, non-negative.
                         It may be an array of shape ``(num_nodes, )`` or a single scalar if the radius is
                         uniform across all nodes.
                         This value is uniform across all ``transformers``.
                         It may be omitted if :attr:`~node_radius` is set.
            show_progress: Whether to print a simple progress bar

        See Also:
            :meth:`~append_properties`

        Raises:
            :class:`ValueError`: if a property with that name is already present
        """
        node_radius = node_radius if node_radius is not None else self.node_radius
        assert node_radius is not None, (
            "parameter node_radius must be set either with the method or the object attribute but is "
            "missing on both"
        )

        with transformer:
            new = transformer.get_transformed_at_nodes(
                self.latitudes_radians, self.longitudes_radians, node_radius, show_progress=show_progress
            )
        self.nodes = self.nodes.join(new)

    def append_properties(
        self,
        transformers: Sequence[BaseTransformer],
        node_radius: Optional[Union[float, ndarray]] = None,
        show_progress: bool = False,
    ) -> None:
        """Append multiple properties at once. This has the benefit of printing a combined progress bar.

        Args:
            transformers: The dimensions/properties that shall be queried for each node
            node_radius: The radius around each node of the area on the globe that it should represent,
                         in meters, non-negative.
                         It may be an array of shape ``(num_nodes, )`` or a single scalar if the radius is
                         uniform across all nodes.
                         This value is uniform across all ``transformers``.
                         It may be omitted if :attr:`~node_radius` is set.
            show_progress: Whether to print a simple progress bar

        See Also:
            :meth:`~append_property`

        Raises:
            ValueError: if a property with any given name is already present
        """
        node_radius = node_radius if node_radius is not None else self.node_radius
        assert node_radius is not None, (
            "parameter node_radius must be set either with the method or the object attribute but is "
            "missing on both"
        )

        for transformer in tqdm(transformers, unit=" transformers", colour="blue", disable=not show_progress):
            self.append_property(transformer, node_radius, show_progress)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, GeoNavigationGraph)
            and NavigationGraph.__eq__(self, other)
            and self.node_radius == other.node_radius
        )
