"""This module provides generic navigation graphs."""

# Typing
from typing import Any
from typing import List
from typing import Optional
from typing import Sized
from typing import Type
from typing import TypeVar

# Mathematics
from numpy import array_equal
from numpy import asarray
from numpy import compress
from numpy import cumsum
from numpy import empty
from numpy import int32
from numpy import logical_and
from numpy import logical_not
from numpy import ndarray

# Scientific
import h5py
from pandas import DataFrame
from pandas import read_hdf


NavigationGraphSubclass = TypeVar("NavigationGraphSubclass", bound="NavigationGraph")


class NavigationGraph(Sized):
    """A generic undirected graph that can be used for navigation.

    It is represented by nodes and their properties as rows in a pandas dataframe, and edges as an array of
    indices of nodes for connections. Additionally, :attr:`~neighbors` array is provided which provides all
    neighbors for all nodes for faster access in graph search.

    Args:
        nodes: the nodes as a dataframe where each row is a node
        edges: the edges of shape ``(number_of_edges, 2)``, where each row contains the indices of two
               neighboring nodes
        neighbours: the neighbors of all nodes with shape ``(number_of_nodes, max_neighbors_per_node)``,
                    where each row contains the indices of all neighbors of the node, filled with ``-1`` at
                    the end
        max_neighbors: the maximum number of neighbors of any node (optional); this can be set to allow for
                       some optimizations (e.g. in the neighbor search)

    Examples:
        This creates a very simple node where ``0`` and ``1`` as well as ``1`` and ``2`` are connected to from
        a small chain.

        >>> nodes = DataFrame(data={'property_1': [1, 2, 3], 'property_2': [10, 20, 30]})
        >>> edges = asarray([[0, 1], [1, 2]])
        >>> graph = NavigationGraph(nodes, edges)
        >>> graph.neighbors
        array([[ 1, -1],
               [ 0,  2],
               [ 1, -1]], dtype=int32)
        >>> len(graph)
        3
        >>> graph.num_edges
        2

    See Also:
        :class:`~pyrate.plan.graph.geo_graph.GeoNavigationGraph`:
            A more specific implementation that references nodes to geographic locations and contains
            useful methods for adding properties from datasets and plotting the graph
    """

    def __init__(
        self,
        nodes: DataFrame,
        edges: ndarray,
        neighbours: Optional[ndarray] = None,
        max_neighbors: Optional[int] = None,
    ) -> None:
        super().__init__()

        self.nodes = nodes

        assert (
            len(edges.shape) == 2 and edges.shape[1] == 2
        ), "the edges must be a 2D-array of shape (number_of_edges, 2)"
        self.edges = edges

        assert neighbours is None or neighbours.shape[0] == len(nodes)
        self._neighbors = neighbours

        assert max_neighbors is None or max_neighbors >= 0, "max_neighbors must be non-negative"
        self.max_neighbors = max_neighbors

    @property
    def neighbors(self) -> ndarray:
        """The list of neighbors of each node, identified by their node index.

        An array of dimension ``(number_of_nodes, max_neighbors_per_node)``, with each row containing the
        indices of the neighbors of the node at that position and the rest of the row filled with ``-1``.

        This might take a short while to be computed for the first time but the result is cached and
        also serialized if present at the point of saving it to disk.
        See :ref:`script-benchmark_graph_neighbor_search` for performance measurements and a link to an issue
        about speeding up this search for neighbors.
        """
        if self._neighbors is not None:
            return self._neighbors

        if self.nodes.empty:  # this requires special case
            self._neighbors = empty((0, 0), dtype=int32)

        else:
            # each row/inner list contains the neighbors of the node at the index
            # and the rest of the row is filled with -1s
            neighbors: List[List[int]] = [[] for _ in range(len(self))]

            for from_node, to_node in self.edges:
                neighbors[from_node].append(to_node)
                neighbors[to_node].append(from_node)

            # calculate length of maximal list
            longest = len(max(neighbors, key=len))
            # make the lists equal in length by filling with -1
            neighbors = [x + [-1] * (longest - len(x)) for x in neighbors]

            self._neighbors = asarray(neighbors, dtype=int32)

        return self._neighbors

    def prune_nodes(self, keep_condition: ndarray) -> None:
        """Only retain the given nodes with their properties and appropriately update all edges and neighbors.

        For example, this should decrease the number of nodes and edges by about 30% when filtering with the
        ``keep_condition`` set to ``my_graph.nodes["elevation_to_sea_level"] < 0.0`` on a graph representing
        earth.

        Args:
            keep_condition: the nodes which to keep as a numpy array with boolean values
        """
        assert keep_condition.shape == (len(self),), "keep condition shape does not match nodes"

        # filter points
        self.nodes = self.nodes[keep_condition]

        # filter edges
        keep_condition_edges = logical_and(keep_condition[self.edges[:, 0]], keep_condition[self.edges[:, 1]])
        self.edges = compress(keep_condition_edges, self.edges, axis=0)
        # then correct the indices that the entries in filtered_edges refer to by subtracting the number of
        # removed entries before each one
        index_shift = cumsum(logical_not(keep_condition)).astype(self.edges.dtype)
        self.edges -= index_shift[self.edges]

        # reset neighbors
        self._neighbors = None

    @staticmethod
    def _serialized_attributes() -> List[str]:
        """The list of attributes that shall be (de)serialized (on top of the nodes and edges)."""
        return ["max_neighbors"]

    def to_disk(self, file_path: str, overwrite_existing: bool = False) -> None:
        """Save the graph to disk. Possibly missing parent directories are automatically created.

        The data is stored in an interoperable `HDF5 <https://docs.h5py.org/en/stable/>`_ file with the
        keys ``nodes``, ``edges`` and optionally ``neighbors``.

        The ``nodes`` are compressed using the default settings of :meth:`pandas.DataFrame.to_hdf`.
        The ``edges`` (and ``neighbors`` if present) are slightly compressed using the library
        `h5py <https://docs.h5py.org/en/stable/high/dataset.html#filter-pipeline>`_ (using GZIP level 4).
        See also `the available options in h5py
        <https://docs.h5py.org/en/stable/faq.html#what-compression-processing-filters-are-supported>`_.

        Args:
            file_path: the path to the file where to store the graph; usually ends with ``.hdf5``
            overwrite_existing: whether to overwrite the file if it already exists; else, this causes an error
                                to be risen

        Raises:
            IOError: when the file cannot be accessed or written to, or it already exists and
                     ``overwrite_existing`` is not set

        See Also:
            :meth:`~from_disk`
        """
        compression_options = {"compression": "gzip", "compression_opts": 4}

        with h5py.File(file_path, "w" if overwrite_existing else "w-") as graph_file:
            graph_file.create_dataset("edges", data=self.edges, **compression_options)
            if self._neighbors is not None:
                graph_file.create_dataset("neighbors", data=self.neighbors, **compression_options)

            # Serialize attributes
            for attribute in self._serialized_attributes():
                graph_file.attrs[attribute] = getattr(self, attribute)

        # pandas automatically chooses an appropriate compression
        self.nodes.to_hdf(file_path, key="nodes", mode="r+", append=True)

    @classmethod
    def from_disk(cls: Type[NavigationGraphSubclass], file_path: str) -> NavigationGraphSubclass:
        """Reads a file from disk.

        Assumes the HDF5-based format compatible to the one created by :meth:`~NavigationGraph`.

        Args:
            file_path: the path to the file where to read the graph from; usually ends with ``.hdf5``

        Raises:
            IOError: when the file cannot be accessed or read from

        Returns:
            The newly loaded navigation graph, which will be of a subclass of :class:`NavigationGraph` if this
            method was called on that class.

        See also:
            :meth:`~to_disk`
        """
        with h5py.File(file_path, "r") as graph_file:
            edges = graph_file["edges"][:]
            neighbors = graph_file["neighbors"][:] if "neighbors" in graph_file else None

            # Deserialize attributes
            attributes = {
                attribute: graph_file.attrs[attribute] for attribute in cls._serialized_attributes()
            }

        nodes = read_hdf(file_path, key="nodes")
        assert isinstance(nodes, DataFrame)

        return cls(nodes, edges, neighbors, **attributes)

    @property
    def num_edges(self) -> int:
        """Returns the number of edges. The number of nodes can be obtained via ``len(graph)``."""
        return self.edges.shape[0]

    def __len__(self) -> int:
        return len(self.nodes)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, NavigationGraph)
            and self.nodes.equals(other.nodes)
            and array_equal(self.edges, other.edges)
            # no need to check array_equal(self.neighbors, other.neighbors) as it is a derived property
            and self.max_neighbors == other.max_neighbors
        )
