"""
cost model code for use in the near-planner
"""

# Dataclass Support
from dataclasses import fields

# Static Typing
from typing import cast
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# Scientific Computing
import numpy as np
from numpy.typing import NDArray

from scipy import linalg

from .evaluated_timing_frame import CollisionData
from .evaluated_timing_frame import EvaluatedTimingFrame

from .timing_frame import TimingFrame

from .obstacle import Obstacle
from .polar_model import PolarModel
from . import utils
# Custom Type for later use
TimingFrameGradients = Tuple[NDArray[np.floating], NDArray[np.floating], NDArray[np.floating]]


class CostModel:
    """Class for evaluating and optimizing cost of timed nautical routes.

    Each route is modeled as a :class:`~.timing_frame.TimingFrame`.

    Args:
        obstacles:
            Dictionary containing :class:`~.obstacle.Obstacle` s to regard in future route evaluations,
            indexed by unique keys.
        polar_model: polar of the boat to simulate
    """

    def __init__(self, obstacles: Dict[str, Obstacle], polar_model: PolarModel) -> None:
        self._obstacles = obstacles
        self._model = polar_model

        self._points_per_obstacle: List[int] = []

        self._obstacles_edges: NDArray[np.floating] = np.array([], dtype=np.float64)
        self._obstacles_points: NDArray[np.floating] = np.array([], dtype=np.float64)
        self._obstacles_speeds: NDArray[np.floating] = np.array([], dtype=np.float64)
        self._obstacles_inner_normals: NDArray[np.floating] = np.array([], dtype=np.float64)

        self._obstacles_squared_edges_length: Optional[NDArray[np.floating]] = np.array([], dtype=np.float64)

        self._update()

    @property
    def obstacles(self) -> Dict[str, Obstacle]:
        """Dictionary of IDs with :class:`~.obstacle.Obstacle`\\ s represented by this :class:`CostModel`."""
        return self._obstacles

    def _update(self) -> None:
        """Updates and recalculates geometric relationships between :class:`~.obstacle.Obstacle`\\ s.

        Warning:
            To be called at any change to :attr:`obstacles` by the internal methods.
        """

        if not self.obstacles:
            return

        self.distance_matrix = self._check_obstacles()

        self._obstacles_points = np.concatenate(
            [obstacle.unique_points for obstacle in self._obstacles.values()], axis=0
        )
        self._obstacles_inner_normals = np.concatenate(
            [o.inner_normals for o in self._obstacles.values()], axis=0
        )
        self._obstacles_edges = np.concatenate([o.edges for o in self._obstacles.values()])
        self._points_per_obstacle = [obstacle.unique_points.shape[0] for obstacle in self._obstacles.values()]

        self._obstacles_speeds = np.array(
            [
                v
                for i, o in enumerate(self._obstacles.values())
                for v in [o.speed] * self._points_per_obstacle[i]
            ]
        )

        # check case if no obstacle
        self._obstacles_speeds = (
            np.array([[]]) if len(self._obstacles_speeds.shape) == 1 else self._obstacles_speeds
        )
        self._obstacles_points = (
            np.array([[]]) if len(self._obstacles_points.shape) == 1 else self._obstacles_points
        )

        self._obstacles_squared_edges_length = np.concatenate(
            [o.squared_edges_length for o in self._obstacles.values()]
        )

    def _check_obstacles(self) -> NDArray[np.floating]:
        obstacle_list = list(self._obstacles.values())
        distance_matrix = np.zeros((len(obstacle_list), len(obstacle_list)))
        for obs1 in range(len(obstacle_list)):
            for obs2 in range(len(obstacle_list)):
                if not obs1 == obs2:
                    distance = np.float64(obstacle_list[obs1].shape.distance(obstacle_list[obs2].shape))
                    distance_matrix[obs1][obs2] = distance
                    if distance == 0:
                        # TODO: join obs1 and obs2
                        pass

        return distance_matrix

    def evaluate(
        self, frame_to_evaluate: TimingFrame, initial_cost: float = 0
    ) -> Tuple[EvaluatedTimingFrame, np.floating]:
        """Evaluates a timed route given in form of a :class:`~.timing_frame.TimingFrame`.

        Upgrades his name to a :class:`~.evaluated_timing_frame.EvaluatedTimingFrame` and also enters a valid
        value for :attr:`~.evaluated_timing_frame.EvaluatedTimingFrame.actual_cost`.

        Args:
            frame_to_evaluate:
                The :class:`~.timing_frame.TimingFrame` to evaluate with regards to models properties
            initial_cost: The base cost of a route just an optional way to vertically scale route cost

        Returns:
            Tuple of the :class:`~.evaluated_timing_frame.EvaluatedTimingFrame` and its actual cost for
            convenience
        """

        if isinstance(frame_to_evaluate, EvaluatedTimingFrame):
            return frame_to_evaluate, frame_to_evaluate.actual_cost

        # Evaluate the frame
        frame_to_evaluate.update_times(self._model)
        data = self._collision_eval(frame_to_evaluate)

        # Wrap in the new object
        evaluated_frame = EvaluatedTimingFrame(
            frame_to_evaluate.route, data, start_time=frame_to_evaluate.start_time
        )
        evaluated_frame.update_times(self._model)

        # Update the actual_cost
        cost = np.float64(initial_cost)
        for key, obstacle in self._obstacles.items():
            min_dist = np.amin(evaluated_frame.data_on_collisions[key].min_dist)
            cost += obstacle.cost(min_dist)
        cost += evaluated_frame.times[-1]
        evaluated_frame.actual_cost = cost

        return evaluated_frame, cost

    def add_obstacle(self, obstacles_to_add: Dict[str, Obstacle]) -> None:
        """Adds an obstacle to the model.

        Args:
            obstacles_to_add:
                The :class:`set` of :class:`~.obstacle.Obstacle`\\ s to add to the model,
                or alternatively a :class:`dict` containing them indexed by IDs
        """
        self._obstacles.update(obstacles_to_add)
        self._update()

    def contains_obstacle(self, obstacle_id_to_check: str) -> bool:
        """Checks if the model describes a given obstacle id.

        This is if and only if they are contained in a held dictionary with the id as key.

        Args:
            obstacle_id_to_check: string key of the obstacle to check if held

        Returns:
            if obstacle is described in this instance
        """
        return obstacle_id_to_check in self._obstacles.keys()

    def delete_obstacle(self, id_s_to_delete: Union[str, List[str]]) -> List[Obstacle]:
        """Takes either a key or a list of keys to delete

        Args:
            id_s_to_delete: list or single key of obstacle(s) to delete from model

        Returns:
            deleted obstacles
        """
        obstacles_deleted = []
        if hasattr(id_s_to_delete, "__iter__"):
            id_s_to_delete = cast(List[str], id_s_to_delete)
            for iter_id in id_s_to_delete:
                deleted_obstacle = self._obstacles.pop(iter_id, None)

                assert deleted_obstacle is not None, "tried to delete a non existent identifier"
                obstacles_deleted.append(deleted_obstacle)
        else:
            id_s_to_delete = cast(str, id_s_to_delete)

            obstacles_deleted.append(self._obstacles[id_s_to_delete])
            del self._obstacles[id_s_to_delete]
        self._update()
        return obstacles_deleted

    def rebase_obstacles(self, obstacles_to_rebase: Dict[str, Obstacle]) -> None:
        """Completely rebases this model on the given obstacles.

        Args:
            obstacles_to_rebase: Dictionary of :class:`~.obstacle.Obstacle`\\ s with keys to rebase on
        """
        self._obstacles = obstacles_to_rebase
        self._update()

    # ---- Route Evaluation

    def _collision_eval(self, frame: TimingFrame) -> Dict[str, CollisionData]:
        """Private collision evaluation logic

        Args:
            frame: The :class:`~.timing_frame.TimingFrame` to be evaluated

        Returns:
            collision data for each obstacle packed in as :class:`~.holders.CollisionData` with id as key
        """
        data = CollisionData()

        if not isinstance(frame, EvaluatedTimingFrame):

            if len(self._obstacles.items()) == 0:
                return {}
            
            seg_points = np.concatenate((utils.shapely_point_to_ndarray(frame.position)[None, :], frame._segment_points), axis=0)

            points_on_route, times, delta_times, speeds = cast(
                Tuple[NDArray[np.floating], NDArray[np.floating], NDArray[np.floating], NDArray[np.floating]],
                (
                    np.array(list(seg_points), dtype=np.float64)[:-1],
                    np.array(list(frame.times[:]), dtype=np.float64)[:-1],
                    np.array(list(np.diff(frame.times)), dtype=np.float64)[:],
                    np.array(list(frame.speeds[:]))[:],
                ),
            )

            #delta_times = np.append(delta_times, np.array([0]), axis=0)[1:]

            edges, data.delta_speed = self.obstacle_time_projection(times, speeds)

            data.delta_points = points_on_route[:, :, None] - edges

            # find time and place of collision
            # normal dist
            data.normal_distances = np.einsum(
                "ix, txi-> ti", self._obstacles_inner_normals, data.delta_points
            )
            # normal speed/vectors
            data.normal_vectors = np.einsum("ix, txi-> ti", self._obstacles_inner_normals, data.delta_speed)

            with np.errstate(divide="ignore", invalid="ignore"):
                data.time_collision = np.nan_to_num(-1 * data.normal_distances / data.normal_vectors)

            with np.errstate(over="ignore"):
                x_coll: NDArray[np.floating] = (
                    np.einsum(
                        "txi, ix-> ti",
                        data.delta_points + data.time_collision[:, None, :] * data.delta_speed,
                        self._obstacles_edges,
                    )
                    / self._obstacles_squared_edges_length
                )

            data.collision_status = (
                (x_coll >= 0)
                * (x_coll <= 1)
                * (data.time_collision >= 0)
                * (data.time_collision <= delta_times[:, None])
            )

            for obs_iter in self._obstacles.values():
                if obs_iter.shape.contains(frame.route.locations[0]) and obs_iter.shape.contains(
                    frame.route.locations[-1]
                ):
                    data.collision_status[0][0] = True

            # determine nearest point between time points
            dv2 = np.einsum("ijk, ijk-> ik", data.delta_speed, data.delta_speed)
            with np.errstate(divide="ignore", invalid="ignore"):
                data.time_min = np.nan_to_num(
                    -1 * np.einsum("ijk, ijk-> ik", data.delta_points, data.delta_speed) / dv2
                )
            
            data.time_min[dv2[:, 0] == 0, :] = 0.0
            data.time_min[(data.time_min < 0) + (data.time_min > delta_times[:, None])] = 0

            # case 3: nearest point between obstacle points
            with np.errstate(divide="ignore", invalid="ignore"):
                x_min = (
                    np.einsum("ijk, kj->ik", data.delta_points, self._obstacles_edges)
                    / self._obstacles_squared_edges_length
                )
            x_min[:, self._obstacles_squared_edges_length == 0] = 0.0
            x_min[(x_min >= 1) + (x_min < 0)] = 0

            # difference vector
            assert self._obstacles is not None
            data.distance_vectors = np.stack(
                (
                    data.delta_points + data.delta_speed * data.time_min[:, None, :],
                    data.delta_points - self._obstacles_edges.T[None, :, :] * x_min[:, None, :],
                ),
                axis=-2,
            )

            data.min_dist = np.amin(linalg.norm(data.distance_vectors, axis=1), axis=-2)
            data.min_dist[data.collision_status] = 0

            data.x_min = x_min
            data.speeds = speeds[:, :, None]

            return self._pack_collision_data(data)

        return frame.data_on_collisions

    def _pack_collision_data(self, data: CollisionData) -> Dict[str, CollisionData]:
        """Packs collision data dictionary created by :meth:`_collision_eval` into a dictionary.

        Done for each :class:`~.obstacle.Obstacle` allows for obstacle specific querying.

        Args:
            data: Dictionary of collision data created by :meth:`_collision_eval`

        Returns:
            Dictionary with obstacle IDs as keys and the :class:`~.holders.CollisionData` as values
        """

        cumulative_points = np.cumsum([0] + self._points_per_obstacle)
        indices = [
            (cumulative_points[i], cumulative_points[i + 1]) for i in range(len(self._points_per_obstacle))
        ]

        return_val: Dict[str, CollisionData] = {}

        for obstacle_iter, ind in zip(self._obstacles.keys(), indices):
            collision_iter: CollisionData = CollisionData()
            for attr in fields(collision_iter):
                data_value_of_attr: np.ndarray = getattr(data, attr.name)
                if data_value_of_attr.shape[-1] > 1:
                    setattr(collision_iter, attr.name, data_value_of_attr.T[ind[0] : ind[1]].T)  # noqa: E203
                else:
                    setattr(collision_iter, attr.name, data_value_of_attr.T[0].T)
            return_val[obstacle_iter] = collision_iter

        return return_val

    def obstacle_time_projection(
        self, times: NDArray[np.floating], speeds: NDArray[np.floating]
    ) -> Tuple[NDArray[np.floating], NDArray[np.floating]]:
        """Projects saved obstacles in their predicted locations at given times.

        Also calculates their vectorized relative speed on route segments relative to each obstacles. For
        the following dimensional information let x denote the cumulated sum of all unique points of all
        obstacles currently stored in the cost model.

        Args:
            times: Time constraints of route with shape ``(number of time points, )``
            speeds: speeds of route ``()``

        Return: A tuple of ``(points, speeds)`` containing a tensor of projected obstacle coordinates at
        each time point of the route``points`` , has dimension ``(number of time points, x, 2)``,
        and ``speeds`` has dimension ``(number of time points, 2, number of obstacles stored, 2)``.

        """

        if len(self._obstacles_points.T.shape) == 1:
            assert len(self._obstacles_speeds.T.shape) == 1, (self._obstacles_points, self._obstacles_speeds)
            # in the case of no obstacles numpy implicitly cast dimensions resulting in an error

            return np.array([[[]]], dtype=np.float64), np.array([[[]]], dtype=np.float64)

        assert len(self._obstacles_speeds.T.shape) == 2, self._obstacles_speeds.shape
        assert len(self._obstacles_points.T.shape) == 2, self._obstacles_points.shape

        projected_points = (
            self._obstacles_points.T[None, :, :] + self._obstacles_speeds.T[None, :, :] * times[:, None, None]
        )
        #print("+"*10)
        #print(times)
        #print(speeds)
        #print("#"*10)
        relative_speeds = speeds[:, :, None] - self._obstacles_speeds.T[None, :, :]
        return projected_points, relative_speeds

    # ---- Gradient Calculation

    def gradients(  # pylint: disable=too-many-locals
        self, frame_s: Union[TimingFrame, List[TimingFrame]]
    ) -> Union[Tuple[List[float], List[NDArray[np.floating]]], Tuple[float, NDArray[np.floating]]]:
        """Calculates the route gradients of the route with respect to full route cost.

        Arguments:
            frame_s: frame(s) to calculate gradients from

        Returns:
            Tuple of cost of taking :class:`~.timing_frame.TimingFrame`\\ (s) and cost derivative of
            :class:`~.timing_frame.TimingFrame`\\ (s)
        """
        if not hasattr(frame_s, "__iter__"):
            frames = [cast(TimingFrame, frame_s)]
        else:
            frames = cast(List[TimingFrame], frame_s)

        evaluated_frames: List[EvaluatedTimingFrame] = []
        for frame in frames:
            frame.update_times(self._model)
            evaluated, _ = self.evaluate(frame)
            evaluated_frames.append(evaluated)

        if not self._obstacles or len(self._obstacles.values()) == 0:
            cost_return, grad_return = [r.cost for r in frames], [r.cost_grad() for r in frames]

            if len(evaluated_frames) == 1:
                assert len(grad_return) == 1
                return cost_return[0], grad_return[0]
            return cost_return, grad_return

        costs_array, obs_gradients = self._calc_gradients(evaluated_frames)
        costs = [c + r.cost for c, r in zip(costs_array, evaluated_frames)]
        gradients = []
        for frame, (grad_subs, grad_times, grad_speeds) in zip(evaluated_frames, obs_gradients):
            assert not np.isnan(grad_times).any(), grad_times
            gradients.append(
                frame.cost_grad(
                    other_cost_dtimes=grad_times, dcost_dspeed=grad_speeds, dcost_dpoints_ext=grad_subs
                )
            )

        if len(evaluated_frames) == 1:
            assert len(gradients) == 1
            return costs[0], gradients[0]
        return costs, gradients

    def _calc_gradients(
        self, frames: List[EvaluatedTimingFrame]
    ) -> Tuple[NDArray[np.floating], Iterable[TimingFrameGradients]]:
        """Calculates cost derivatives in respect to several aspects of a list of frames

        Args:
            frames: frames to calculate Gradients

        Returns: Tuple of cumulative cost and an iterable that yields a tuple of gradients. The first
            entry in the gradient tuple describes the route gradient w.r.t. the locations of the stored
            obstacles. The second entry describes the gradient w.r.t the simulated times  and the third
            gradient is calculated w.r.t. the obstacle speeds under an assumption of constant speed. Each
            gradient should've the shape ``(number of route segments, 2)``. The length of the iterable
            is the same as the given list of frames.
        """

        # pylint: disable-msg=too-many-locals
        # this is ok due to it being already split into several hierarchical methods and i see no way to
        # reduce local variables without butchering visibility
        # shorten it

        # evaluate the attributes at the lowest distance, needed for gradient calculation
        array_shape = (len(frames), len(self._obstacles))
        time_seg, obst_seg, min_dist, grad_distances, t_min, cost = [np.zeros(array_shape) for _ in range(6)]
        speeds, dist_vect = [np.zeros((*array_shape, 2)) for _ in range(2)]
        for i, frame in enumerate(frames):

            for j, (obstacle_key, obstacle) in enumerate(self._obstacles.items()):
                data = frame.data_on_collisions[obstacle_key]

                # pylint: disable-msg=unbalanced-tuple-unpacking
                # this is safe due to how data is created
                frame_index, obstacle_index = np.unravel_index(
                    np.argmin(data.min_dist, axis=None), data.min_dist.shape
                )
                min_dist[i, j] = data.min_dist[frame_index, obstacle_index]

                assert not (np.isinf(min_dist).any()), data

                t_min[i, j] = data.time_min[frame_index, obstacle_index]
                cost[i, j], grad_distances[i, j] = obstacle.distance_gradient(min_dist[i, j])

                # if obstacle.soft and min_dist[i, j] == 0:
                # TO (BJK): add soft obstacles (this is a leftover from old code)
                #    cost[i, j] += obstacle.duration_cost.cost(self._duration_eval(route, obstacle))
                # grad_distances[i, j] = 0

                if min_dist[i, j] == 0:
                    min_dist[i, j] = 1e-3
                time_seg[i, j] = frame_index
                obst_seg[i, j] = obstacle_index

                speeds[i, j, :] = data.speeds[frame_index, :]

                assert data.distance_vectors.shape[2] == 2, "case not handled"

                distance_index = data.distance_vectors[frame_index, :, :, obstacle_index]
                dist_vect[i, j, :] = min(distance_index[:, 0], distance_index[:, 1], key=linalg.norm)

        # chain rule for derivation
        assert not (np.isnan(grad_distances).any()), grad_distances
        assert not (np.isnan(min_dist).any()), min_dist
        assert not (np.isnan(dist_vect).any()), dist_vect

        dcost_d_distvec = np.nan_to_num((grad_distances / min_dist))[:, :, None] * dist_vect

        assert not (np.isnan(dcost_d_distvec).any()), min_dist

        dcost_d_deltapoints = dcost_d_distvec
        dcost_d_speed = dcost_d_distvec * t_min[:, :, None]
        dcost_d_times = np.sum(dcost_d_deltapoints * (-speeds), axis=-1)

        assert not (np.isnan(dcost_d_times).any()), (speeds, grad_distances, min_dist, dist_vect)

        return (
            cast(NDArray[np.floating], cost.sum(axis=1)),
            self._split_grads((dcost_d_deltapoints, dcost_d_times, dcost_d_speed), time_seg, frames),
        )

    def _split_grads(
        self,
        gradients: Tuple[NDArray[np.floating], NDArray[np.floating], NDArray[np.floating]],
        time_seg: NDArray[np.floating],
        frames: List[EvaluatedTimingFrame],
    ) -> Generator[TimingFrameGradients, None, None]:
        """Split combined gradients into separate gradients for each different routes.

        Args:
            gradients: gradients (grad points, grad times, grad velocity) each gradient of the shape
                ``(number of obstacles, number of route segments, 2)``
            time_seg: indices of minimal distance time segment
            frames: frames of the gradients

        Yields: Gradients split per route as a Tuple. The first entry in the gradient tuple describes the
            route gradient w.r.t. the locations of the stored obstacles. The second entry describes the
            gradient w.r.t the simulated times  and the third gradient is calculated w.r.t. the obstacle
            speeds under an assumption of constant speed. Each gradient should've the shape
            ``(number of route segments, 2)``
        """
        for i, frame in enumerate(frames):
            frame_length = frame.times.shape[0]

            grad_subs, grad_times, grad_speeds = (
                np.zeros((frame_length - 1, 2)),
                np.zeros(frame_length),
                np.zeros((frame_length, 2)),
            )
            for j, _ in enumerate(self._obstacles):
                # (leftovers from old code may safe time if soft obstacles to be added)
                # if False:
                # if obst.soft and frame_index.data_on_collision[on].any():
                #    g_subs, g_times, g_speeds = self.duration_grad(frame_index, on, obst)
                #    grad_times += g_times
                #    grad_speeds += g_speeds
                #    grad_subs += g_subs
                # else:

                time_segment = int(time_seg[i, j])
                if time_segment:
                    grad_subs[time_segment - 1] += gradients[0][i, j, :]
                grad_times[time_segment] += gradients[1][i, j]
                grad_speeds[time_segment, :] += gradients[2][i, j, :]

            yield (
                cast(NDArray[np.floating], grad_subs),
                cast(NDArray[np.floating], grad_times),
                cast(NDArray[np.floating], grad_speeds),
            )
