"""Contains evaluated timing frames."""

# Static Typing
from typing import cast
from typing import Dict
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import numpy.typing as npt

# Scientific Computing
import numpy as np

# Geometry
from pyrate.plan.geometry.location import CartesianLocation
from pyrate.plan.geometry.route import CartesianRoute

from .holders import CollisionData
from .timing_frame import TimingFrame


class EvaluatedTimingFrame(TimingFrame):
    """Class used as a wrapper for evaluated :class:`~.timing_frame.TimingFrame`.

    It is guaranteed that instances of this class posses a :attr:`data_on_collision` Evaluation.
    For :class:`~.timing_frame.TimingFrame` instances upgraded by a CostModel simulated times are also
    guaranteed.

    Args:
        route: route simulated by the :class:`EvaluatedTimingFrame`
        data_on_collisions: data on possible collision to encapsulate
        start_time: start time to start route at

    Attributes:
        actual_cost: accumulated cost of time taken, distance to obstacles and their speed
    """

    def __init__(
        self, route: CartesianRoute, data_on_collisions: Dict[str, CollisionData], start_time: float = 0
    ) -> None:
        self._data_on_collisions = data_on_collisions
        self._revoked = False
        self.actual_cost = np.float64(0.0)
        super().__init__(route, start_time)

    def __str__(self) -> str:
        return f"Evaluated{super().__str__()}"

    def revoke(self, new_position: CartesianLocation) -> TimingFrame:
        """Revokes the `~.timing_frame.TimingFrame`.

        This results in this instance becoming invalid
        and returns a new non evaluated :class:`~.timing_frame.TimingFrame` with adapted first point.

        Args:
            new_position: new first position of the revoked :class:`~.timing_frame.TimingFrame`

        Returns:
            stripped :class:`~.timing_frame.TimingFrame`
        """
        self.position = new_position
        new_frame: TimingFrame = TimingFrame(self.route, start_time=self.start_time)
        self._revoked = True
        return new_frame

    @property
    def valid(self) -> bool:
        """True iff :class:`~.timing_frame.TimingFrame` was already evaluated and no collision was found"""
        if self._data_on_collisions is not None and not self._revoked:
            return not any(o.collision_status.any() for k, o in self._data_on_collisions.items())
        return False

    @property
    def imminent_collision(self) -> Tuple[str, float]:
        """Collision time for imminent collisions

        Raises:
            AssertionError: If attribute is called before TimingFrame has been evaluated

        Return:
            tuple of the obstacle id and time of the next collision
        """
        assert self.data_on_collisions is not None
        temp = {}

        for key, obstacle_data in self.data_on_collisions.items():
            if not obstacle_data.collision_status.any():
                temp[key] = [np.inf]
            else:
                temp[key] = (obstacle_data.time_collision + self._times[1:, None])[
                    obstacle_data.collision_status
                ].min()

        return cast(Tuple[str, float], temp.get(cast(str, min(temp.items(), key=lambda x: x[1]))))

    @property
    def collision_times(self) -> Dict[str, float]:
        """Times for collisions with each obstacle

        Raises:
            AssertionError: If attribute is called before TimingFrame hasn't yet been evaluated

        Return:
            dictionary that maps an obstacle id to the time of collision if route is taken
        """
        assert self.data_on_collisions is not None
        temp = {}
        for key, obstacle_data in self.data_on_collisions.items():
            if not obstacle_data.collision_status.any():
                temp[key] = np.inf
            else:
                temp[key] = (obstacle_data.time_collision + self._times[:-1, None])[
                    obstacle_data.collision_status
                ].min()
        return cast(Dict[str, float], temp)

    @property
    def collision_segment(self) -> Dict[str, Optional[npt.NDArray[np.floating]]]:
        """Segments where collision happens as a dictionary with obstacle ids as key.

        Raises:
            AssertionError: If attribute is called before TimingFrame hasn't yet been evaluated

        Return:
            Dictionary mapping obstacle IDs to index of route segment where collision is predicted
        """
        assert self.data_on_collisions is not None
        temp: Dict[str, Optional[npt.NDArray[np.floating]]] = {}
        for key, obstacle_data in self.data_on_collisions.items():
            if not obstacle_data.collision_status.any():
                temp[key] = None
            else:
                temp[key] = np.argmax(obstacle_data.collision_status.sum(axis=1))  # type: ignore
        return temp

    @property
    def minimal_dist(
        self,
    ) -> Union[Set[np.floating], Dict[str, Tuple[np.floating, Optional[npt.NDArray[np.floating]]]]]:
        """Dictionary of minimal distances per obstacle ID.

        Return:
            If not yet evaluated ``{0}``, else a mapping from obstacle ID to ``(distance_x, distance_y)``.
        """
        if self.data_on_collisions is None:
            return {np.float64(0.0)}
        temp: Dict[str, Tuple[np.floating, Optional[npt.NDArray[np.floating]]]] = {}
        for key, obstacle_data in self.data_on_collisions.items():
            if obstacle_data.collision_status.any():
                temp[key] = np.float64(0.0), None
            else:
                temp[key] = (
                    np.float64(obstacle_data.min_dist.min()),
                    np.argmin(obstacle_data.min_dist.min(axis=1)).astype(np.float64),
                )  # type: ignore
        return temp

    @property
    def data_on_collisions(self) -> Dict[str, CollisionData]:
        """Information about collisions with each obstacle, index by obstacle keys."""
        return self._data_on_collisions
