"""collection of dataclasses used in the planning backend and node-service of the near-planner node"""

# Dataclass Support
from dataclasses import dataclass

from typing import Optional

# Scientific Computing
import numpy as np

# Static Typing
import numpy.typing as npt


@dataclass
class CollisionData:
    """Holds information on possible future collisions or collision avoidance with a single obstacle.

    The stored information is not intended to be directly read from this dataclass, but to be stored
    and post-processed internally in an :class:`~planning_backend.timing_frame.EvaluatedTimingFrame`
    and corresponding information be read through its properties.
    """

    collision_status: npt.NDArray[np.bool_] = np.array([], dtype=np.bool_)
    """array, whose entries describe if a  segment i of a frame has a collision with an obstacle """
    min_dist: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, whose entries describe the minimal distances at each time steps to the obstacle"""
    normal_distances: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, whose entries describe the normal distances of obstacle towards the boat on each segment"""
    normal_vectors: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, whose entries describe the normal vectors of each obstacle at each time step"""
    time_collision: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, whose entries describe the time of a future collision with an object"""
    time_min: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, whose entries describe the time of the first collision"""
    x_min: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, whose entries describe the place of the first collision"""
    delta_speed: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, describing the relative speed of route segments towards the obstacle"""
    delta_points: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, describing relative position of route towards obstacle on each time step"""
    speeds: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, describing relative speeds towards obstacle"""
    distance_vectors: npt.NDArray[np.floating] = np.array([], dtype=np.float64)
    """array, describing the vectorized distances towards the given obstacle"""


@dataclass
class EstimationParameters:
    """Holds estimation parameters for finding first initial routes.

    This class is mainly used by instances of :class:`~.optimizer.Optimizer` while searching an initial guess
    for a valid :class:`~.timing_frame.TimingFrame` that later can be optimized.
    """

    first_try_construction: bool = True
    """if first to try deterministic construction by obstacle hulls"""

    max_len_relative: float = 0.1
    """Maximum fraction of the length of a new edge to explore, relative to the distance to the goal"""
    max_count: int = 300
    """maximal number of potential nodes are explored before resulting in a failure"""
    p_goal: float = 0.2
    """probability the goal is chosen as the next node"""


@dataclass
class OptimizationParameters:
    """Holds parameters for running :class:`~.optimizer.Optimizer`."""

    estimation_parameters: EstimationParameters
    """parameters used for initial exploration"""
    n_samples: int = 25
    """maximum number of initial samples that will be selected, optimized and the results compared"""
    inital_stepsize: float = 0.1
    """fraction of the direct distance to goal which will be used as the initial step size for optimization"""
    n_iter_grad: int = 500
    """maximum number of steps used by the gradient descent algorithm"""
    n_break: int = 10
    """maximum number of initial samples which will be selected to be optimized and compared"""
    adaptive_step_size: bool = True
    """if to use adaptive step size during gradient descent"""
    prune: bool = True
    """if to prune the resulting routes"""
    verbose: bool = False
    """if to print additional information"""
    overwrite_grad_lim: Optional[float] = None
    """optional parameter to overwrite to cut of the gradient at each optimization step"""
    adaptive_data_rate_penalty_on_loss_sign_change: float = 2.5
    """penalty applied to lossful sign changes during adaptive step size optimization"""
    adaptive_data_rate_general_regulization_factor: float = 1.5
    """regularization positive and negative used to adapt step size based on whether cost in/decreases"""
