"""
code for the Obstacle
"""

# Static Typing
from typing import cast
from typing import Tuple

from numpy.typing import NDArray

# Scientific Computing
import numpy as np

from scipy import linalg

# Geometry
from pyrate.plan.geometry.polygon import CartesianPolygon

from .cost_functions import CostFunction
from .cost_functions import CostFunctionLinear

class Obstacle:
    """An obstacle with a polygon shape and vector speed.

    Wrapper class around a :class:`~pyrate.plan.geometry.polygon.CartesianPolygon`. Assumes constant speed.

    Args:
        polygon: Polygon shape obstacle shall be based on
        speed: 2D speed vector describing the obstacle's movement in ``(east, north)`` direction.
        cost_function: The cost function that shall apply to this :class:`~.obstacle.Obstacle`
    """

    def __init__(
        self,
        polygon: CartesianPolygon,
        speed: NDArray[np.floating],
        cost_function: CostFunction = CostFunctionLinear(),
    ) -> None:
        self.shape = polygon
        self.speed = speed if speed is not None else np.zeros(2)
        self.cost_function = cost_function

    def __call__(self, time: float) -> "Obstacle":
        """Transposes Obstacle to a given time.

        Args:
            time: time to transpose to

        Returns:
            A copy of this object transposed to given time
        """
        projection = self.project_at_time(time)
        cartesian_poly = CartesianPolygon.from_numpy(projection)
        return Obstacle(cartesian_poly, self.speed, cost_function=self.cost_function)

    def distance_gradient(self, min_dist: float) -> Tuple[np.floating, np.floating]:
        """Calculates the distance gradient.

        Args:
            min_dist: minimal distance to calculate distance gradient from

        Returns:
            derivative of cost with regards to minimal distance
        """
        return self.cost_function.cost_grad(min_dist)

    def cost(self, min_distance: float) -> np.floating:
        """Calculates the cost of the obstacle in respect to the minimal distance.

        Args:
            min_distance: minimal distance to base the cost of the obstacle on

        Returns:
            cost of taking a route in the minimal distance of this obstacle
        """
        return self.cost_function.cost(min_distance)

    def project_at_time(self, time: float) -> NDArray[np.floating]:
        """Projects the obstacle shape at a given time.

        Args:
            time: time to project the obstacle shape to

        Returns:
            numpy.ndarray with coordinates of the projected shape
        """
        return cast(NDArray[np.floating], self.shape.to_numpy() + self.speed * time)

    @property
    def unique_points(self) -> NDArray[np.floating]:
        """only unique points of polygon shape

        Returns:
            numpy-array of unique points that make up the polygon
        """
        return cast(NDArray[np.floating], self.shape.to_numpy()[:-1, :])

    @property
    def edges(self) -> NDArray[np.floating]:
        """method for getting edges in a numpy array

        Returns:
            numpy-array of polygon edges
        """
        temp = self.shape.to_numpy()
        return cast(NDArray[np.floating], (np.concatenate((temp[1:, :], temp[:1, :])) - temp)[:-1, :])

    @property
    def squared_edges_length(self) -> NDArray[np.floating]:
        """method for getting the squared length of polygon edges in a numpy array

        Returns:
            numpy-array of squared lengths of polygon edges
        """
        return cast(NDArray[np.floating], linalg.norm(self.edges, axis=1) ** 2)

    @property
    def edges_length(self) -> NDArray[np.floating]:
        """method for getting edges length in a numpy array

        Returns:
            numpy-array of lengths of polygon edges
        """
        return cast(NDArray[np.floating], linalg.norm(self.edges, axis=1))

    @property
    def inner_normals(self) -> NDArray[np.floating]:
        """method for getting normals in numpy-array

        Returns:
            numpy-array of edge normals
        """
        diffs = self.edges
        normals = np.concatenate((diffs[:, 1:2], -diffs[:, 0:1]), axis=1)
        with np.errstate(divide="ignore"):
            _ = cast(NDArray[np.floating], normals / linalg.norm(normals, axis=1, keepdims=True))
        return _
