"""
This module implements the ``ImageLine`` used by ``ObstacleLocator`` to construct the horizon in an image.
"""

from __future__ import annotations

# Standard library
from math import atan2
from math import cos
from math import sin
from math import sqrt
from math import tan

# Typing
from typing import cast
from typing import Tuple

# Numpy
import numpy
from numpy import arange
from numpy import arctan
from numpy import diff
from numpy import linspace
from numpy import ndarray
from numpy import pi


class ImageLine:

    """Represents a line in the image.

    The line can be constructed using
    either points, a height and an angle or polar coordinates using the respective ``from_X`` methods.

    Args:
        image_shape: Shape of the image ``(width, height)``
        x_y_coordinates: x- and y-coordinates of both endpoints as ``((x1, x2), (y1, y2))``
        height: Height of the line in the image space
        angle: Angle of the line in the image
    """

    def __init__(
        self,
        image_shape: Tuple[int, int],
        x_y_coordinates: Tuple[Tuple[int, int], Tuple[int, int]],
        height: int,
        angle: float,
    ):
        self.image_width, self.image_height = image_shape
        self.x_y_coordinates = x_y_coordinates
        self.height = height
        self.angle = angle

        self.x_points = arange(x_y_coordinates[0][0], x_y_coordinates[0][1] + 1).astype(int)
        self.y_points = numpy.round(linspace(*x_y_coordinates[1], num=len(self.x_points))).astype(int)

    @classmethod
    def from_points(
        cls, image_shape: Tuple[int, int], points: Tuple[Tuple[int, int], Tuple[int, int]]
    ) -> ImageLine:
        """Construct an ``ImageLine`` from the two endpoints of the line (in the image space).

        Args:
            image_shape: Shape of the image ``(width, height)``
            points: End points of the line ``((x1, y1), (x2, y2))``

        Returns:
            New ``ImageLine`` from two points
        """

        x_coordinates = cast(Tuple[int, int], tuple(p[0] for p in points))
        y_coordinates = cast(Tuple[int, int], tuple(p[1] for p in points))

        height = numpy.sum(y_coordinates) / 2
        angle = atan2(diff(y_coordinates), diff(x_coordinates))

        return cls(
            image_shape=image_shape,
            x_y_coordinates=(x_coordinates, y_coordinates),
            height=height,
            angle=angle,
        )

    @classmethod
    def from_height_angle(
        cls, image_shape: Tuple[int, int], relative_height: float, angle: float
    ) -> ImageLine:
        """Construct an ``ImageLine`` from its relative height and angle.

        Note: This method of creation might create end_points and indices values that are
        out of bound of the image if the angle of the represented line is too large.

        Args:
            image_shape: Shape of the image ``(width, height)``
            relative_height: Relative height of the line (based on ``floor(image_shape[0] / 2)``)
            angle: Angle of the line

        Returns:
            New ``ImageLine`` from height and angle
        """

        x_coordinates = (0, image_shape[0])
        diff_height = int(tan(angle) * image_shape[0] / 2)
        height = int(relative_height + image_shape[1] / 2)
        y_coordinates = (height - diff_height, height + diff_height)

        return cls(
            image_shape=image_shape,
            x_y_coordinates=(x_coordinates, y_coordinates),
            height=height,
            angle=angle,
        )

    @classmethod
    def from_polar(cls, image_shape: Tuple[int, int], radius: float, alpha: float) -> ImageLine:
        # pylint: disable=too-many-locals
        """Construct an ``ImageLine`` from its polar representation.

        The line is represented by its distance from the coordinate origin (radius) and the
        line rotation angle in radians.

        Note: This method of creation might create end_points and indices values that are
        out of bound of the image if the angle of the represented line is too large.

        Args:
            image_shape: Shape of the image ``(width, height)``
            radius: Distance from the coordinate origin
            alpha: Line rotation angle in radians

        Returns:
            New ``ImageLine`` from the polar representation
        """

        cos_alpha, sin_alpha = cos(alpha), sin(alpha)
        # line_end pixels
        x_coordinates = (0, image_shape[0])
        y_coordinates = (int(radius / sin_alpha), int((radius - cos_alpha * image_shape[0]) / sin_alpha))
        # pixels of a line
        x_points: ndarray = arange(0, image_shape[0])
        y_points: ndarray = numpy.round((radius - cos_alpha * x_points) / sin_alpha).astype(int)

        angle = alpha - pi / 2  # of horizon
        d_2, alpha_0 = sqrt(image_shape[1] ** 2 + image_shape[0] ** 2) / 2, arctan(
            image_shape[1] / image_shape[0]
        )
        height_rel = radius - d_2 * cos(alpha - alpha_0)  # over image center
        height = int(height_rel + image_shape[1] // 2)

        line = cls(
            image_shape=image_shape,
            x_y_coordinates=(x_coordinates, y_coordinates),
            height=height,
            angle=angle,
        )
        line.x_points = x_points
        line.y_points = y_points
        return line

    @property
    def indices(self) -> Tuple[ndarray, ndarray]:
        """Returns pixel coordinates of the line on the image

        Returns:
            two ``ndarray``s containing every x and y coordinate of pixels the line covers
        """

        return self.x_points, self.y_points

    @property
    def end_points(self) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """Returns the two end points defining the line

        Returns:
            x, y coordinates of both end points as in ``((x1, x2), (y1, y2))``
        """

        return (self.x_y_coordinates[0][0], self.x_y_coordinates[1][0]), (
            self.x_y_coordinates[0][1],
            self.x_y_coordinates[1][1],
        )
