"""
This module implements the representation of an obstacle or object in an image
(modelled as a bounding box), used by ``ObstacleLocator``.
"""

# Typing
from typing import Tuple


class ImageRectangle:

    """Represents an object's bounding box in the image plane as a rectangular bounding box.

    Args:
        rectangle: x and y coordinates, width and height of the rectangle
        offset: position offset (e.g. due to subimage used for detection)
        angle: rotation offset (due to rotate detection)
    """

    def __init__(
        self, rectangle: Tuple[int, int, int, int], offset: Tuple[int, int] = (0, 0), angle: float = 0
    ) -> None:

        self.x_rel, self.y_rel, self.width, self.height = rectangle
        self.offset = offset
        self._x_coord, self._y_coord = self.x_rel + offset[0], self.y_rel + offset[1]
        self.angle = angle

    def rectangle_to_corner(self, offset: bool = True) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """Returns bounding box corners for OpenCV drawing functions.

        Args:
            offset: If True, return coordinates offset by the offset specified when
            creating the `ImageRectangle`. If False, ignore the offset.

        Returns:
            Coordinates of the upper left and lower right points of the rectangle ``(x1, y1), (x2, y2)``
        """

        if offset:
            x_res, y_res = self._x_coord, self._y_coord
        else:
            x_res, y_res = self.x_rel, self.y_rel
        return (x_res, y_res), (x_res + self.width, y_res + self.height)

    @property
    def bottom_center(self) -> Tuple[int, int]:
        """Returns the bottom center coordinates of the bounding box.

        Returns:
            x, y coordinates of the bottom center pixel
        """

        return self._x_coord + self.width // 2, self._y_coord + self.height
