"""
This module implements a mechanism to construct the horizon line in an image,
as well as (in the future) find obstacles in the image and return their locations.
"""

# Typing
from typing import List
from typing import Tuple

# Numpy
import numpy
from numpy import arctan
from numpy import ndarray
from numpy import pi
from numpy import radians

# Scientific
from cv2 import blur
from cv2 import Canny
from cv2 import HoughLines
from cv2 import INTER_AREA
from cv2 import medianBlur
from cv2 import resize

# Auxiliary modules
from .image_line import ImageLine


class ObstacleLocator:  # pylint: disable=too-few-public-methods

    """Canny based horizon extraction and (in the future) obstacle detection class.

    Args:
        image_width: Width of images in pixel
        image_height: Height of images in pixel
    """

    def __init__(self, image_width: int, image_height: int) -> None:
        # Precompute common image attributes
        self.image_width, self.image_height = image_width, image_height
        self._center = image_width // 2, image_height // 2
        self._min_angle = arctan(self.image_width / self.image_height)

    def _preprocess(self, image: ndarray, box_blur_size: int = 90, median_blur_size: int = 3) -> ndarray:
        """Preprocesses an image, reduces its resolution and smooths it.

        Args:
            image: Image to preprocess
            box_blur_size: The size of the box blur filter is calculated as
            `image_height // box_blur_size`. Only modify default value with prior testing.
            median_blur_size: Size of the median filter aperture. Must be
            odd and greater than 1; only modify default value with prior testing.

        Returns:
            Preprocessed image
        """

        image = resize(image, (self.image_width, self.image_height), interpolation=INTER_AREA)
        blur_size = self.image_height // box_blur_size
        image = blur(image, (blur_size, blur_size))
        image = medianBlur(image, median_blur_size)
        return image

    def _detect_lines(  # pylint: disable=too-many-arguments
        self,
        image: ndarray,
        n_max: int,
        min_line_length: float,
        canny_threshold: int,
        angle_resolution: float = 0.2,
    ) -> Tuple[List[ImageLine], List[int]]:
        """Detect lines in an image using the Canny and Hough filter

        Args:
            image: Image where lines should be detected in
            n_max: Max number of lines that should be returned
            min_line_length: Line length threshold
            canny_threshold: Threshold for the hysteresis procedure
            angle_resolution: Distance resolution of the accumulator

        Returns:
            Detected lines, their lengths and the dissimilarity of the regions separated by each line
        """

        # extract lines in image
        edges = Canny(image, canny_threshold, canny_threshold * 2)
        lines = HoughLines(
            edges,
            1,
            radians(angle_resolution),
            int(self.image_width * min_line_length),
            min_theta=self._min_angle,
            max_theta=pi - self._min_angle,
        )
        if lines is None:
            return [], []
        lines = [
            ImageLine.from_polar((self.image_width, self.image_height), *line[0]) for line in lines[:n_max]
        ]
        line_lengths = self._evaluate_lines(edges, lines)

        return lines, line_lengths

    @staticmethod
    def _evaluate_lines(
        edges: ndarray,
        lines: List[ImageLine],
    ) -> List[int]:
        """Evaluate the length and separation of image by each line.

        Args:
            image: The image the line is drawn on
            edges: Edges in image
            lines: ``ImageLine``s that shall be evaluated

        Returns:
            A tuple containing each line's length and the dissimilarity of the regions separated by the line
        """

        # number of points on line
        votes = [numpy.sum(edges[line.y_points, line.x_points] > 0).astype(int) for line in lines]
        return votes

    def detect_horizon(
        self, img: ndarray, n_max: int = 3, min_visibility: float = 0.2, canny_threshold: int = 35
    ) -> Tuple[List[ImageLine], List[int]]:
        """Detects possible horizon line(s).

        Args:
            img: Image to analyze
            n_max: Max number of lines to extract
            min_visibility: Minimum length of horizon line to be considered
            canny_threshold: Canny threshold for edge detection

        Returns:
            Horizon lines (in image space shrunk to ``(self._w, self._h)`` as set in :meth:`__init__`!),
            their respective lengths, and
            their separation factors (dissimilarity of the two regions of the image
            separated by that line)
        """

        img = self._preprocess(img)
        return self._detect_lines(img, n_max, min_visibility, canny_threshold)
