"""This module asserts correct runtime behaviour of the pyrate.sense.filter subpackage."""

# Python standard library
from datetime import timedelta
from typing import cast

# Test environment
from unittest import TestCase

# Mathematics
from numpy import array
from numpy import eye
from numpy import Inf
from numpy.linalg import norm
from numpy import ndarray
from numpy import vstack

# Hypothesis testing
from hypothesis import given
from hypothesis import settings

# Package under test
from pyrate.common.math import Gaussian
from pyrate.sense.filters import ExtendedKalman
from pyrate.sense.filters import Kalman
from pyrate.sense.filters import UnscentedKalman

# Helpers
from pyrate.common.testing.strategies.dynamic_system import linear_model
from pyrate.common.testing.strategies.dynamic_system import nonlinear_model

# Flags from Pyrate
from pyrate.common.testing import IS_EXTENDED_TESTING


class TestKalman(TestCase):

    """Test for correct runtime behaviour of Kalman filters pyrate.sense.filter."""

    # In this context, we reproduce a common filter notation
    # pylint: disable=invalid-name

    @given(linear_model())
    @settings(max_examples=1000 if IS_EXTENDED_TESTING else 10, deadline=timedelta(seconds=1.0))
    def test_kalman(self, model):
        """Assert the correct functionality of the standard Kalman filter."""

        # Unpack generated model
        estimate, F, B, H, Q, R, measurements, inputs = model

        # Initialize filter
        kalman = Kalman(F=F, estimate=estimate, H=H, Q=Q, R=R, B=B, keep_trace=True)

        # Apply the Kalman filter a few times
        for i, measurement in enumerate(measurements):
            kalman.predict(u=inputs[i])
            kalman.correct(z=measurement)

        # Assert correct tracing with Kalman
        self.assertIsNotNone(kalman.predictions, "Kalman filter did not keep trace of predictions")
        self.assertIsNotNone(kalman.estimates, "Kalman filter did not keep trace of estimates")
        self.assertEqual(
            len(kalman.predictions.index), len(measurements), "Kalman filter has not traced all predictions"
        )
        self.assertEqual(
            len(kalman.estimates.index), len(measurements), "Kalman filter has not traced all estimates"
        )

        # Apply the Kalman filter with a callable H
        callable_H = lambda dummy: H  # noqa: E731
        kalman.H = callable_H
        kalman.correct(z=measurements[0], dummy=None)

    @given(nonlinear_model())
    @settings(max_examples=1000 if IS_EXTENDED_TESTING else 10, deadline=timedelta(seconds=1.0))
    def test_extended_kalman(self, model):
        """Assert the correct functionality of the extended Kalman filter."""

        # Unpack generated model
        estimate, f, _, Jf, h, _, Jh, Q, R, measurements = model

        # Initialize filter
        extended = ExtendedKalman(F=Jf, f=f, estimate=estimate, H=Jh, h=h, Q=Q, R=R, keep_trace=True)

        # Apply the Kalman filter a few times
        for z in measurements:
            extended.predict()
            extended.correct(z)

        # Assert correct tracing with Kalman
        self.assertIsNotNone(extended.predictions, "Extended Kalman filter did not keep trace of predictions")
        self.assertIsNotNone(extended.estimates, "Extended Kalman filter did not keep trace of estimates")
        self.assertEqual(
            len(extended.predictions.index),
            len(measurements),
            "Extended Kalman filter has not traced all predictions",
        )
        self.assertEqual(
            len(extended.estimates.index),
            len(measurements),
            "Extended Kalman filter has not traced all estimates",
        )

    @given(nonlinear_model())
    @settings(max_examples=1000 if IS_EXTENDED_TESTING else 10, deadline=timedelta(seconds=1.0))
    def test_unscented_kalman(self, model):
        """Assert the correct functionality of the extended Kalman filter."""

        # Unpack generated model
        estimate, f, _, _, h, _, _, Q, R, measurements = model

        # Initialize filter
        unscented = UnscentedKalman(f=f, estimate=estimate, h=h, Q=Q, R=R, keep_trace=True)

        # Apply the Kalman filter a few times
        for z in measurements:
            unscented.predict()
            unscented.correct(z)

        # Assert correct tracing with Kalman
        self.assertIsNotNone(
            unscented.predictions, "Uncented Kalman filter did not keep trace of predictions"
        )
        self.assertIsNotNone(unscented.estimates, "Uncented Kalman filter did not keep trace of estimates")
        self.assertEqual(
            len(unscented.predictions.index),
            len(measurements),
            "Uncented Kalman filter has not traced all predictions",
        )
        self.assertEqual(
            len(unscented.estimates.index),
            len(measurements),
            "Uncented Kalman filter has not traced all estimates",
        )

    @staticmethod
    def test_estimation():
        """Assert that the filter estimates tend towards the true state over time."""

        # Define model of a constant value
        F = H = Q = R = eye(1)

        # Initial belief
        estimate = Gaussian(vstack([0.0]), eye(1))

        # Initialize filters
        kalman = Kalman(F=F, estimate=estimate, H=H, Q=Q, R=R)
        extended = ExtendedKalman(
            F=lambda _: F,
            f=lambda x: cast(ndarray, x),
            estimate=estimate,
            H=lambda _: H,
            h=lambda x: cast(ndarray, x),
            Q=Q,
            R=R,
        )
        unscented = UnscentedKalman(
            f=lambda x: cast(ndarray, x),
            estimate=estimate,
            h=lambda x: cast(ndarray, x),
            Q=Q,
            R=R,
        )

        # Apply the Kalman filter a few times
        true_state = array([20.0])
        previous_kalman_error = Inf
        previous_extended_error = Inf
        previous_unscented_error = Inf
        for _ in range(10):
            # Check error going down for Kalman
            kalman.predict()
            kalman.correct(z=true_state)
            error = norm(kalman.estimate.x - true_state).item()  # Convert from numpy scalar to Python float
            assert (
                error < previous_kalman_error or error == 0
            ), "Kalman estimate did not get better over time."
            previous_kalman_error = error

            # Check error going down for EKF
            extended.predict()
            extended.correct(z=true_state)
            error = norm(extended.estimate.x - true_state).item()
            assert error < previous_extended_error or error == 0, "EKF estimate did not get better over time."
            previous_extended_error = error

            # Check error going down for UKF
            unscented.predict()
            unscented.correct(z=true_state)
            error = norm(unscented.estimate.x - true_state).item()
            assert (
                error < previous_unscented_error or error == 0
            ), "UKF estimate did not get better over time."
            previous_unscented_error = error
